import tensorflow as tf
import numpy as np
from ...graph import Graph, NodeKeys


class Normalizer(Graph):

    def __init__(self, name, input_tensors=None, **config):
        super().__init__(name, **config)
        self.register_node(NodeKeys.INPUT, input_tensor)

        self.register_main_task(self.__normalization)
        self.register_task('denorm', self.__denormalization)
        self.register_task('norm', self.__normalization)

    def __unified_input_tensors(self, input_tensors):
        pass

    def _normalization_kernel(self, feeds):
        raise NotImplementedError

    def _denormalization_kernel(self, feeds):
        raise NotImplementedError

    def __normalization(self, feeds):
        return self._normalization_kernel(self.get_feed_dict(feeds))

    def __denormalization(self, feeds):
        return self._denormalization_kernel(self.get_feed_dict(feeds))


class FixWhite(Normalizer):
    method_name = 'fix_white'

    def __init__(self, name, input_tensor, **config):
        super().__init__(name, input_tensor, **config)

    def _normalization_kernel(self, feeds):
        with tf.name_scope('normalization'):
            return (feeds - self.c['mean']) / self.c['std']

    def _denormalization_kernel(self, feeds):
        with tf.name_scope('denormalization'):
            return feeds * self.c['std'] + self.c['mean']


class MeanStdWhite(Normalizer):
    method_name = 'mean_std_white'

    def __init__(self, name, input_tensor, **config):
        super().__init__(name, input_tensor, **config)

    def _normalization_kernel(self, feeds):
        with tf.name_scope('normalization'):
            mean = tf.reduce_mean(feeds)
            maxb = tf.reduce_max(feeds)
            minb = tf.reduce_min(feeds)
            return (feeds - self.c['mean']) / self.c['std']


class SelfMinMax(Normalizer):
    method_name = 'self_min_max'

    def __init__(self, name, input_tensor, **config):
        super().__init__(name, input_tensor, **config)

    def _normalization_kernel(self, feeds):
        if isinstance(feeds, tf.Tensor):
            with tf.name_scope('normalization'):
                rmin = tf.reduce_min(feeds)
                rmax = tf.reduce_max(feeds)
                data = (feeds - rmin) / (rmax - rmin)
        else:
            rmin = np.min(feeds)
            rmax = np.max(feeds)
            data = (feeds - rmin) / (rmax - rmin)
        return {'min': rmin, 'max': rmax, 'data': data}

    def _denormalization_kernel(self, feeds):
        with tf.name_scope('denormalization'):
            return feeds['data'] * (feeds['max'] - feeds['min']) + feeds['min']


class SelfMeanStd(Normalizer):
    method_name = 'self_mean_std'

    def __init__(self, name, input_tensor, **config):
        super().__init__(name, input_tensor, **config)

    def _normalization_kernel(self, feeds):
        if isinstance(feeds, tf.Tensor):
            with tf.name_scope('normalization'):
                mean, stdv = tf.nn.moments(feeds)
                data = (feeds - mean) / stdv
        else:
            mean = np.mean(feeds)
            stdv = np.std(feeds)
            data = (feeds - mean) / stdv
        return {'mean': mean, 'std': stdv, 'data': data}

    def _denormalization_kernel(self, feeds):
        with tf.name_scope('denormalization'):
            return feeds['data'] * feeds['std'] + feeds['mean']


class ReduceSum(Normalizer):
    def __init__(self, name='normalizer/reduce_sum', **config):
        super().__init__(name, **config)



from .base import Service
from .web import create as create_web
from .web import read as read_web
from .web import read_all as read_all_web
from .web import update as update_web
from .web import delete as delete_web


def create_py(task_json):
    return Service.create(task_json)


def read_py(tid):
    return Service.read(tid)


def read_all_py():
    return Service.read_all()


def update_py(task_json):
    return Service.update(task_json)


def delete_py(tid):
    return Service.delete(tid)



from ..path import Path


class Configs:
    def __init__(self, base_path, dct_or_config_viewer):
        if isinstance(dct_or_config_viewer, Configs):
            base_path = dct_or_config_viewer.base / str(Path(base_path))
            self.data = dct_or_config_viewer.data
        else:
            self.data = dct_or_config_viewer
        self.base = Path(base_path)

    def __get_by_keys(self, keys):

        while result is None:
            result = self
        for k in keys:
            result =

    def get(self, key):
        keys = self._unified_and_processing_input_key(key)

    def __getitem__(self, key):
        return self.get(key)


class ConfigsSetter:
    def __init__(self, config_dict):
        self._data = config_dict

    def update(self, path, dct):
        target = self._get_item(Path(path).parent())
        target[Path(path).basename].update(dct)

    def join(self, path, dct):
        target = self._get_item(Path(path).parent())
        for k in target[Path(path).basename]:

        target[Path(path).basename] = dct

    def _get_item(self, path):
        result = self
        for k in Path(path).parts():
            result = result[k]
        return result

    def __getitem__(self, key):
        from ..code import DXPYTypeError
        if not key in self._data:
            self._data[key] = dict()
            return ConfigsSetter(self._data[key])
        elif isinstance(self._data[key], dict):
            return ConfigsSetter(self._data[key])
        else:
            raise DXPYTypeError(dict, self._data[key],
                                'Value of Key {}'.format(key))

    def __setitem__(self, key, value):
        self._data[key] = value



import yaml
import rx
import logging
from enum import Enum
logger = logging.getLogger(name=__name__)




class Tags:
    undifined = '<undifined>'


def with_default(x, default=None):
    return x if x is not None else default


def add_yaml_support_enum(cls, tag):
    fmt = 'YAML support added for {cls} with tag: {tag}.'
    logger.info(fmt.format(cls=cls, tag=tag))

    def representer(dumper, data):
        return dumper.represent_scalar(tag, data.name)
    yaml.add_representer(cls, representer)

    def constructor(loader, node):
        value = loader.construct_scalar(node)
        for v in cls:
            if v.name == value:
                return v
        raise ValueError(
            "Unkown value {value} of type {cls}.".format(value=value, cls=cls))
    yaml.add_constructor(tag, constructor)


def add_yaml_support(cls, tag):
    if issubclass(cls, Enum):
        add_yaml_support_enum(cls, tag)


def notebook_logging_support():
    logger = logging.getLogger()
    logger.setLevel(logging.INFO)
    handler = logging.StreamHandler(sys.stderr)
    logger.handlers = [handler]
    logger.info('test')


def urlf(host, port, url, method="http"):
    return "{method}://{host}:{port}{url}".format(method=method, host=host, port=port, url=url)


def rx_block(o):
    return (o.suscribe_on(rx.concurrency.ThreadPoolScheduler())
            .to_blocking().first())



from .directory.model import Directory
from .file.model import File
from .path.model import Path
from .directory import service as ds
from .file import service as fs
from fs.osfs import OSFS


def mv(sor, tar, overwrite=True):
    if isinstance(sor, Directory):
        ds.mv(sor, tar, overwrite)
    else:
        fs.mv(sor, tar, overwrite)


def cp(sor, tar, overwrite=True):
    if isinstance(sor, Directory):
        ds.cp(sor, tar, overwrite)
    else:
        fs.cp(sor, tar, overwrite)


def rm(sor):
    if isinstance(sor, Directory):
        ds.rm(sor)
    else:
        fs.rm(sor)


class EnsureFS:
    def __init__(self, fs_or_path, default_filesystem=OSFS):
        if isinstance(fs_or_path, str):
            self.fs = default_filesystem(Path(fs_or_path).abs)
            self.need_close = True
        else:
            self.fs = fs_or_path
            self.need_close = False

    def __enter__(self):
        return self.fs

    def __exit__(self, type, value, trackback):
        if self.need_close:
            self.fs.close()


def launch_web_ui(host, port, version=0.1, debug=True):
    from flask import Flask, url_for
    from flask_cors import CORS, cross_origin
    from flask_restful import Api
    from .web import add_apis
    app = Flask(__name__)
    CORS(app)
    api = Api(app)
    add_apis(api, '/api/v{0}'.format(version))

    @app.route("/site-map")
    def site_map():
        import json
        links = []
        for rule in app.url_map.iter_rules():
            if rule.endpoint == 'static':
                continue
            links.append(
                (url_for(rule.endpoint, path='%252Ftmp%252Ftest'), rule.endpoint))
        return json.dumps(links, indent=4, separators=(',', ':'))
    app.run(host, port, debug=debug)



import tensorflow as tf


def mean_square_error(label, data):
    with tf.name_scope('mean_squared_error'):
        return tf.sqrt(tf.reduce_mean(tf.square(label - data)))


def rmse(label, data):
    with tf.name_scope('rmse'):
        rmse = tf.sqrt(tf.reduce_mean(tf.square(label - data)))
    return rmse


def psnr(label, data):
    with tf.name_scope('psnr'):
        r = rmse(label, data)
        maxv = tf.reduce_max(label)
        minv = tf.reduce_min(label)
        sca = 255.0 / (maxv - minv)
        ln = (label - minv) * sca
        tn = (data - minv) * sca
        rmv = rmse(ln, tn)
        value = 10.0 * tf.log((255.0**2.0) / (rmv**2.0)) / tf.log(10.0)
    return value



class PhantomSpec:
    def __init__(self, shape):
        self._shape = shape

    @property
    def shape(self):
        return self._shape


class Phantom2DSpec(PhantomSpec):
    ndim = 2

    def __init__(self, shape):
        super().__init__(shape)


class Phantom3DSpec(PhantomSpec):
    ndim = 3

    def __init__(self, shape):
        super().__init__(shape)



import click

def get_main_net(net_name):
    from dxpy.learn.net.zoo.srms.train import main as srms
    if net_name == "srms":
        return srms

@click.command()
@click.option('--config', '-c', type=str, help='configs .yml filename')
@click.option('--net', '-n', type=str, help='name of main net', default=None)
def train(net, config):    
    main_net_train = get_main_net(net)
    main_net_train()

@click.command()
@click.option('--config', '-c', type=str, help='configs .yml filename', default='dxln.yml')
@click.option('--name', '-n', type=str, help='cluster name of dataset in config file, e.g. cluster/dataset/task0.')
def dataset(name, config):
    from .dataset import dataset_dist
    from dxpy.learn.utils.general import load_yaml_config
    load_yaml_config(config)
    dataset_dist(name)

@click.command()
@click.option('--name', '-n', type=str, help='name of network')
@click.option('--config', '-c', type=str, help='configs .yml filename', default='dxln.yml')
@click.option('--job_name', '-j', type=str)
@click.option('--task_index', '-t', type=int)
@click.option('--run', '-r', type=str, help='run task', default='train')
def main(name, job_name, task_index, run, config):
    from dxpy.learn.utils.general import load_yaml_config
    load_yaml_config(config)
    if name == 'sin':
        from .zoo.sin.main import main
        main(run, job_name, task_index)
    elif name == 'srms':
        from .zoo.srms.main import main
        main(run, job_name, task_index)
    else:
        raise ValueError("Unknown name {} for dxpy.ln.main CLI.".format(name))



from dxpy.filesystem import Path
from treelib import Node, Tree


class PathTree:
    def __init__(self):
        self.tree = Tree()
        self.create_node()

    @classmethod
    def _resolve_path(cls, name, path_parent):
        if name is None:
            if path_parent is None:
                return None, Path('/').abs
            else:
                raise TypeError("Name is None while path_parent is not None")
        else:
            if path_parent is None:
                return Path('/').abs, (Path('/') / name).abs
            else:
                return Path(path_parent).abs, (Path(path_parent) / name).abs

    def create_node(self, name=None, path_parent=None, data=None):
        path_parent, path_current = self._resolve_path(name, path_parent)
        if path_parent is not None and self.get_node(path_parent) is None:
            self.create_node(Path(path_parent).name,
                             Path(path_parent).father, None)
        return self.tree.create_node(name, path_current, parent=path_parent, data=data)

    def get_node(self, path):
        return self.tree.get_node(path)
    

    def get_data(self, path):
        return self.get_node(path).data

    def sub_tree(self, path):
        return self.tree.sub_tree(path)



from collections import UserDict
from dxpy import serialization
from dxpy.filesystem import Path
from ._exceptions import NotDictError


class DXDict(UserDict):
    yaml_tag = '!dxdict'

    def __init__(self, *args, default_dict=None, **kwargs):
        super(__class__, self).__init__(*args, **kwargs)
        if default_dict is not None:
            self.default_dict = DXDict(default_dict)
        else:
            self.default_dict = None

    def __getitem__(self, key):
        if key in self.data:
            return self.data[key]
        elif self.default_dict is not None:
            return self.default_dict[key]
        return None

    def keys(self):
        from itertools import chain
        if self.default_dict is not None:
            return chain(self.data.keys(), self.default_dict.keys())
        else:
            return self.data.keys()

    def apply_default(self, default_dict):
        if self.default_dict is None:
            return DXDict(self.data, default_dict=default_dict)
        else:
            return DXDict(self.data, default_dict=self.default_dict.apply_default(default_dict))


serialization.register(DXDict)

# from .trees import PathTree


# class TreeDict(PathTree):
#     yaml_tag = '!treedict'

#     def __init__(self, *args, **kwargs):
#         super(__class__, self).__init__()

#     def compile(self):
#         self._push_dict()

#     def _push_dict(self, path=None, dct=None):
#         if dct is None:
#             dct = DXDict()
#         if path is None:
#             path = '/'
#         if self.get_data(path) is None:
#             self.get_node(path).data = DXDict(dct)
#             new_dict = self.get_node(path).data
#         else:
#             new_dict = self.get_data(path).apply_default(dct)
#         self.get_node(path).data = new_dict
#         nodes = self.tree.children(path)
#         for n in nodes:
#             self._push_dict(n.indentifier, self.get_node(path).data)

#     def __getitem__(self, path):
#         return self.get_data(path)

#     def add_dict(self, name, parent, dct):
#         self.create_node(name, parent, data=DXDict(dct))


# TODO: Add yaml support.
class TreeDict(UserDict):
    """
    A dict with tree struct:
        1. Supports three different access method:
            1. a['key1']['key2']
            2. a['/key1/key2']
            3. a.get(['key1','key2'])
            4. a.get('key1')
        2. Autogenerate for missing layers
            If there is no a['name1'] dict while setting a['name1']['name2'],
            an empty TreeDict will be created on a['name1']
        3. Inherence values
            If we are accessing a['name1']['name2'], but there is not 'name2' key in a['name1'],
            the dict will try to find if there is 'name2' in a (root).
        4. KeyError only happens when need a dict but got a object which is not dict,
           For those keys are not defined in the TreeDict, return a empty TreeDict object.
    """
    yaml_tag = '!dxdict'

    def __init__(self, dct=None, fa=None):
        """

        Inputs:
            - dct: dict like object, dict, Userdict or TreeDict,
            - fa: father node, object of TreeDict,

        Raises:
            None

        """
        # if dct is not None and not isinstance(dct, (dict, UserDict)):
        #     raise NotDictError(type(dct))
        super(__class__, self).__init__(dct)
        self.fa = fa
        for k in self.data:
            self._unfied_dict_element(k, False)

    def get(self, key):
        keys = self._unified_and_processing_input_key(key)
        return self._get(keys)

    def __getitem__(self, key):
        return self.get(key)

    def set(self, key, value):
        keys = self._unified_and_processing_input_key(key)
        self._set(keys, value)

    def __setitem__(self, key, value):
        self.set(key, value)

    def empty(self):
        return len(self.data) == 0

    def publish(self):
        for kv in self.data:
            if not isinstance(self.data[kv], TreeDict):
                for kd in self.data:
                    if isinstance(self.data[kd], TreeDict):
                        self.data[kd]._publish_element(kv, self.data[kv])
        for k in self.data:
            if isinstance(self.data[k], TreeDict):
                self.data[k].publish()

    def _publish_element(self, key, value):
        if key in self.data and not isinstance(self.data[key], TreeDict):
            return
        if key in self.data and isinstance(self.data[key], TreeDict) and len(self.data[key]) > 0:
            return
        self.data[key] = value

    def _ensure_mid_layers(self, keys):
        if len(keys) > 1:
            if not keys[0] in self.data:
                self.data[keys[0]] = TreeDict(fa=self)
            else:
                self._unfied_dict_element(keys[0], True)

    @classmethod
    def _unified_keys(cls, key_or_keys_or_path):
        if isinstance(key_or_keys_or_path, (list, tuple)):
            return tuple(key_or_keys_or_path)
        if isinstance(key_or_keys_or_path, str):
            result = tuple(Path(key_or_keys_or_path).parts)
            if len(result) > 0 and result[0] == '/':
                result = result[1:]
            return result
        if isinstance(key_or_keys_or_path, int):
            return tuple(str(key_or_keys_or_path))
        else:
            raise TypeError(
                "Key not supported {}.".format(key_or_keys_or_path))

    def _update_child_dct(self, dct, key=None):
        from ruamel.yaml.comments import CommentedMap
        if isinstance(dct, TreeDict):
            if dct.fa is None:
                dct.fa = self
            return dct
        result = None
        if isinstance(dct, (dict, UserDict)):
            result = TreeDict(dct, fa=self)
        # if isinstance(dct, (list, tuple)):
        #     result = dict()
        #     for i, v in enumerate(dct):
        #         result[str(i)] = v
        #     result = TreeDict(result, fa=self)
        if isinstance(dct, CommentedMap):
            result = dict()
            for k in dct:
                result[k] = dct[k]
            result = TreeDict(result, fa=self)
        if result is not None and key is not None:
            self.data[key] = result
        return result

    def _unfied_dict_element(self, key, required=False):
        from ._exceptions import KeyNotDictError
        result = self._update_child_dct(self.data[key], key)
        if result is None and required:
            raise KeyNotDictError(key, type(self.data[key]))
        return result

    def _unified_and_processing_input_key(self, key_or_keys_or_path):
        keys = self._unified_keys(key_or_keys_or_path)
        self._ensure_mid_layers(keys)
        return keys

    def _get_value_by_key_from_inherence(self, key):
        if self.fa is None:
            result = None
        else:
            result = self.fa._get_value_by_key(key)

        if result is None:
            self.data[key] = TreeDict()
            result = self.data[key]
        return result

    def _get_value_by_key(self, key):
        if key in self.data:
            return self.data[key]
        return self._get_value_by_key_from_inherence(key)

    def _get(self, keys):
        if len(keys) == 0:
            raise ValueError(
                "TreeDict requires positve len(keys), probably using key = None?")
        if len(keys) == 1:
            return self._get_value_by_key(keys[0])
        result = self._unfied_dict_element(keys[0], True)
        return result.get(keys[1:])

    def _set(self, keys, value):
        if len(keys) == 1:
            self.data[keys[0]] = value
            self._unfied_dict_element(keys[0], False)
        else:
            self.data[keys[0]].set(keys[1:], value)


class FileSystemDict(UserDict):
    def __init__(self, dct):
        super(__class__, self).__init__(dct)



from dxpy.core.path import Path


class TensorSpec:
    def __init__(self, path='data', shape=None, dtype=None, chunks=None):
        self._path = Path(path)
        self.shape = shape
        self.dtype = dtype
        self.chunks = chunks

    @property
    def dataset(self):
        return self._path.basename

    @property
    def groups(self):
        return self._path.parts()[:-1]

    def get_group(self, h5py_file, *, restrict=False):
        ds = h5py_file
        for g in self.groups:
            if restrict:
                ds = ds[g]
            else:
                ds = ds.require_group(g)
        return ds

    def create_dateset(self, h5py_file, data=None):
        gp = self.get_group(h5py_file)
        if data is None:
            return gp.create_dateset(self.dataset, shape=self.shape, dtype=self.dtype, chunks=chunks)
        else:
            return gp.create_dataset(self.dataset, data=data)

    def get_dataset(self, h5py_file):
        gp = self.get_group(h5py_file)
        return gp[self.basename]

class Tensor:
    def __init__(self, ):
        pass



import tensorflow as tf
from ...base import Net, Graph, NodeKeys
from ....model.cnn.super_resolution import SuperResolutionMultiScalev2, SRKeys


class SRNetKeys:
    RES_INF = 'res_inf'
    RES_ITP = 'res_itp'
    LOSS_ITP = 'loss_itp'


from dxpy.configs import configurable
from ....config import config

# class SRMultiScale(Net):
#     """
#     Required inputs:
#     'input/image2x'|...|'input/image8x'
#     'label/image1x', ..., 'label/image4x'
#     """
#     @configurable(config, with_name=True)
#     def __init__(self, inputs, name='network', nb_gpu=2, nb_down_sample=3, **kw):
#         super().__init__(name, inputs, nb_gpu=nb_gpu, nb_down_sample=nb_down_sample, **kw)


#     def summary_items(self):
#         from ....train.summary import SummaryItem
#         from ....scalar import global_step
#         result = {
#             'loss':     SummaryItem(self.nodes[NodeKeys.EVALUATE]),
#             'global_step': SummaryItem(global_step())}
#         max_down_sample = self.param('nb_down_sample')
#         result.update({'img{}x'.format(2**max_down_sample):
#                        SummaryItem(self.nodes['input/image{}x'.format(2**max_down_sample)])})
#         result.update({'inf': SummaryItem(self.nodes[NodeKeys.INFERENCE])})
#         result.update(
#             {'itp': SummaryItem(self.nodes['outputs/{}'.format(SRKeys.INTERP)])})
#         result.update({'label': SummaryItem(
#             self.nodes['outputs/{}'.format(SRKeys.ALIGNED_LABEL)])})
#         result.update({'res_inf': SummaryItem(
#             self.nodes['outputs/{}'.format(SRNetKeys.RES_INF)])})
#         result.update({'res_itp': SummaryItem(
#             self.nodes['outputs/{}'.format(SRNetKeys.RES_ITP)])})

#         if 'outputs/'+SRKeys.POI_LOSS in self.nodes is not None:
#             result.update({'poi_loss': SummaryItem(self.noddes['outputs/'+SRKeys.POI_LOSS])})
#             result.update({'mse_loss': SummaryItem(self.noddes['outputs/'+SRKeys.MSE_LOSS])})
#         return result
#         # def _get_node(self, node_type, down_sample_ratio, source=None):
#         #     if source is None:
#         #         source = self.nodes
#         #     name = "{}/image{}x".format(node_type, 2**down_sample_ratio)
#         #     return source[name]

#     def _post_kernel_post_outputs(self):
#         self.register_node(NodeKeys.EVALUATE,
#                            self.nodes['outputs/{}'.format(NodeKeys.EVALUATE)])
#         super()._post_kernel_post_outputs()

#     def _kernel(self, feeds):

#         from ....model.tensor import MultiGPUSplitor, PlaceHolder
#         from ....utils.general import device_name
#         from ....model.image import align_crop
#         from dxpy.collections.dicts import swap_dict_hierarchy
#         mgs = MultiGPUSplitor(nb_gpu=self.param('nb_gpu'))
#         feeds_gpus = mgs(feeds)
#         first_part = list(feeds_gpus.values())[0]
#         with tf.name_scope('cpu_placeholders'):
#             cpu_placeholders = {k: PlaceHolder(
#                 first_part[k]).as_tensor() for k in first_part}
#         with tf.device(device_name('cpu')):
#             model = SuperResolutionMultiScalev2(name='model', inputs=cpu_placeholders)
#             model()
#         result_gpus = {}
#         for i, k in enumerate(mgs.part_names()):
#             with tf.device(device_name('gpu', i)):
#                 result_gpus[k] = model(feeds_gpus[k])
#         with tf.device(device_name('cpu')):
#             with tf.name_scope('merge'):
#                 with tf.name_scope('inference'):
#                     infer = tf.concat([result_gpus[k][NodeKeys.INFERENCE]
#                                        for k in mgs.part_names()], axis=0)
#                 with tf.name_scope('label'):
#                     label = tf.concat([result_gpus[k][SRKeys.ALIGNED_LABEL]
#                                        for k in mgs.part_names()], axis=0)
#                 with tf.name_scope('interp'):
#                     interp = tf.concat([result_gpus[k][SRKeys.INTERP]
#                                         for k in mgs.part_names()], axis=0)
#                 with tf.name_scope('residual'):
#                     res_inf = tf.abs(label - infer)
#                     res_itp = tf.abs(label - interp)
#                 losses = [result_gpus[k][NodeKeys.LOSS]
#                           for k in mgs.part_names()]
#                 if SRKeys.POI_LOSS in result_gpus[mgs.part_names()[0]]:
#                     losses_poi = [result_gpus[k][SRKeys.POI_LOSS]
#                                   for k in mgs.part_names()]
#                     losses_mse = [result_gpus[k][SRKeys.MSE_LOSS]
#                                   for k in mgs.part_names()]
#                 else:
#                     losses_poi = None
#                     losses_mse = None

#                 with tf.name_scope('total_loss'):
#                     loss = tf.add_n(losses)
#                     if losses_poi is not None:
#                         loss_poi = tf.add_n(losses_poi)
#                         loss_mse = tf.add_n(losses_mse)
#                     else:
#                         loss_poi = None
#                         loss_mse = None

#         result = {NodeKeys.INFERENCE: infer,
#                   SRKeys.ALIGNED_LABEL: label,
#                   SRKeys.INTERP: interp,
#                   SRNetKeys.RES_INF: res_inf,
#                   SRNetKeys.RES_ITP: res_itp,
#                   NodeKeys.LOSS: losses,
#                   NodeKeys.EVALUATE: loss}
#         if loss_poi is not None:
#             result.update({SRKeys.MSE_LOSS: loss_mse,
#                            SRKeys.POI_LOSS: loss_poi})
#         return result

class SRMultiScale(Net):
    """
    Required inputs:
    'input/image2x'|...|'input/image8x'
    'label/image1x', ..., 'label/image4x'
    """
    @configurable(config, with_name=True)
    def __init__(self, inputs, name='network', nb_gpu=2, nb_down_sample=3, new_trainer=False, **kw):
        super().__init__(name, inputs, nb_gpu=nb_gpu,
                         nb_down_sample=nb_down_sample, new_trainer=new_trainer, **kw)

    def summary_items(self):
        from ....train.summary import SummaryItem
        from ....scalar import global_step
        result = {'loss':     SummaryItem(self.nodes[NodeKeys.EVALUATE])}
        result.update({'loss_itp': SummaryItem(
            self.nodes['outputs/{}'.format(SRNetKeys.LOSS_ITP)])})
        max_down_sample = self.param('nb_down_sample')
        result.update({'img{}x'.format(2**max_down_sample):
                       SummaryItem(self.nodes['input/image{}x'.format(2**max_down_sample)])})
        result.update({'inf': SummaryItem(self.nodes[NodeKeys.INFERENCE])})
        result.update(
            {'itp': SummaryItem(self.nodes['outputs/{}'.format(SRKeys.INTERP)])})
        result.update({'label': SummaryItem(
            self.nodes['outputs/{}'.format(SRKeys.ALIGNED_LABEL)])})
        result.update({'res_inf': SummaryItem(
            self.nodes['outputs/{}'.format(SRNetKeys.RES_INF)])})
        result.update({'res_itp': SummaryItem(
            self.nodes['outputs/{}'.format(SRNetKeys.RES_ITP)])})

        if 'outputs/' + SRKeys.POI_LOSS in self.nodes is not None:
            result.update({'poi_loss': SummaryItem(
                self.noddes['outputs/' + SRKeys.POI_LOSS])})
            result.update({'mse_loss': SummaryItem(
                self.noddes['outputs/' + SRKeys.MSE_LOSS])})
        return result
        # def _get_node(self, node_type, down_sample_ratio, source=None):
        #     if source is None:
        #         source = self.nodes
        #     name = "{}/image{}x".format(node_type, 2**down_sample_ratio)
        #     return source[name]

    def _post_kernel_post_outputs(self):
        self.register_node(NodeKeys.EVALUATE,
                           self.nodes['outputs/{}'.format(NodeKeys.EVALUATE)])
        # super()._post_kernel_post_outputs()
        from dxpy.learn.train.trainer_2 import Trainer
        if NodeKeys.LOSS in self.nodes:
            self.register_node(NodeKeys.TRAINER,
                               Trainer(self.name / 'trainer', self.nodes[NodeKeys.LOSS]))

    def _kernel(self, feeds):
        from ....model.tensor import MultiGPUSplitor, PlaceHolder
        from ....utils.general import device_name
        from ....model.image import align_crop
        from dxpy.learn.model.losses import mean_square_error
        from dxpy.collections.dicts import swap_dict_hierarchy
        result = SuperResolutionMultiScalev2(name='model', inputs=feeds)()
        with tf.name_scope('residuals'):
            res_inf = tf.abs(
                result[SRKeys.ALIGNED_LABEL] - result[NodeKeys.INFERENCE])
            res_itp = tf.abs(
                result[SRKeys.ALIGNED_LABEL] - result[SRKeys.INTERP])
        if NodeKeys.LOSS in result:
            loss_itp = mean_square_error(
                result[SRKeys.ALIGNED_LABEL], result[SRKeys.INTERP])
            result.update({SRNetKeys.LOSS_ITP: loss_itp})
        result.update({SRNetKeys.RES_INF: res_inf})
        result.update({SRNetKeys.RES_ITP: res_inf})
        result.update({NodeKeys.EVALUATE: result[NodeKeys.LOSS]})
        return result
    # def _kernel_multiscale(self, image_lowest, image_labels):
    #     """
    #     Returns:
    #         images_infer, cropped inferenced images (different scale)
    #         loss (total loss)
    #     """
    #     rep = None
    #     ipt = image_lowest
    #     losses = []
    #     for i in reversed(range(self.param(nb_down_sample))):
    #         with tf.variable_scope('sr2x_{}'.format(i)):
    #             ipt, loss, rep = self._kernel_sr2x(ipt, image_labels[i], rep,
    #                                                name='srb_{}'.format(i))
    #             losses.append(loss * ((0.5)**i))
    #     with tf.name_scope("loss"):
    #         loss = tf.add_n(losses)
    #     return ipt, loss



class AssertEqualSubtestsForFunc:
    def __init__(self, test_case, func):
        self.tests = []
        self.test_case = test_case
        self.func = func

    def add(self, inputs, outputs):
        self.tests.append((inputs, outputs))

    def run(self):
        for ipt, opt in self.tests:
            self.test_case.assertEqual(self.func(ipt), opt)


def debug_msg(*args, **kwargs):
    from inspect import currentframe, getframeinfo
    frameinfo = getframeinfo(currentframe().f_back)
    print('DEBUG MSG FROM:', frameinfo.filename, frameinfo.lineno)
    print(*args, **kwargs)
    print('DEBUG MSG END========================================')



#!/home/hongxwing/anaconda3/bin/python
import click


class CLI(click.MultiCommand):
    commands = {'code': None,
                'task': None,
                'batch': None,
                'run': None,
                'ln': None}

    def __init__(self):
        super(__class__, self).__init__(name='dxl', help='DXL CLI tools.')

    def list_commands(self, ctx):
        return sorted(self.commands.keys())

    def get_command(self, ctx, name):
        from ..task import cli as task
        from ..code import cli as code
        from ..batch.api.cli import batch
        from ..run import run_cli
        from ..learn.run.cli import main as learn 
        if name in self.commands:
            if self.commands[name] is None:
                mapping = {
                    'code': code,
                    'task': task,
                    'batch': batch,
                    'run': run_cli,
                    'ln': learn 
                }
                self.commands[name] = mapping.get(name)
        return self.commands.get(name)


cli = CLI()



import numpy as np
def assert_ndim(tensor, valid_dims, tensor_name=None):
    if isinstance(valid_dims, int):
        valid_dims = [valid_dims]
    if tensor.ndim not in valid_dims:
        prefix = "Tensor {} ".format(tensor_name) if tensor is not None else "Tensor "
        msg = "{} is required to have dims in {}, got {}.".format(prefix,
                                                                  valid_dims, tensor.ndim)
        raise ValueError(msg.format())


def assert_batch_ndim(tensors, valid_dims, tensor_name=None):
    if isinstance(valid_dims, int):
        valid_dims = [valid_dims]
    if tensor.ndim not in valid_dims:
        prefix = "Tensor {} ".format(tensor_name) if tensor is not None else "Tensor "
        msg = "{} is required to have dims in {}, got {}.".format(prefix,
                                                                  valid_dims, tensor.ndim)
        raise ValueError(msg.format())



from dxpy.configs import configurable
from dxpy.learn.config import config
from dxpy.learn.train.summary_2 import SummaryWriter
from .main import NodeKeys, SRKeys, SRNetKeys
import tensorflow as tf


class SRSummaryWriter(SummaryWriter):
    @configurable(config, with_name=True)
    def __init__(self, network, name='summary', **kw):
        summary_tensors = {
            'input_lr': network['inputs/input/image{}x'.format(2**network.param('nb_down_sample'))],
            'label': network['outputs/aligned_label'],
            'interp': network['outputs/interp'],
            'loss': network[NodeKeys.LOSS],
            'infer': network['outputs/inference'],
            'learning_rate': network['trainer']['learning_rate']['value']
        }
        nb_runs = {
            'loss': 32,
            'mes_inf': 32,
            'mse_itp': 32,
            'mse_inf_to_itp_ratio': 32,
        }
        super().__init__(name=name, tensors=summary_tensors, nb_runs=nb_runs, **kw)

    def _processing(self, tensors):
        from dxpy.learn.model.metrics import mse
        with tf.name_scope("processing"):
            result = {k: tensors[k] for k in tensors}
            res_inf = tf.abs(result['label'] - result['infer'])
            res_itp = tf.abs(result['label'] - result['interp'])
            dif_inf_itp = tf.abs(result['infer'] - result['interp'])
            result['res_inf']= res_inf
            result['res_itp']= res_itp
            result['dif_inf_itp']= dif_inf_itp
            result['mse_inf'] = mse(result['label'], result['infer'])
            result['mse_itp'] = mse(result['label'], result['interp'])
            result['mse_inf_to_itp_ratio'] = result['mse_inf'] / \
                result['mse_itp']
        return super()._processing(result)

class SRSummaryWriter_v2(SummaryWriter):
    @configurable(config, with_name=True)
    def __init__(self, network, tensors, name='summary', **kw):
        summary_tensors = {
            'input_lr': tensors['input/image{}x'.format(2**network.param('nb_down_sample'))],
            'label': tensors['aligned_label'],
            'interp': tensors['interp'],
            'loss': tensors[NodeKeys.LOSS],
            'infer': tensors['inference'],
            'learning_rate': network['trainer']['learning_rate']['value']
        }
        nb_runs = {
            'loss': 32,
            'mes_inf': 32,
            'mse_itp': 32,
            'mse_inf_to_itp_ratio': 32,
        }
        super().__init__(name=name, tensors=summary_tensors, nb_runs=nb_runs, **kw)

    def _processing(self, tensors):
        from dxpy.learn.model.metrics import mse
        with tf.name_scope("processing"):
            result = {k: tensors[k] for k in tensors}
            res_inf = tf.abs(result['label'] - result['infer'])
            res_itp = tf.abs(result['label'] - result['interp'])
            dif_inf_itp = tf.abs(result['infer'] - result['interp'])
            result['res_inf']= res_inf
            result['res_itp']= res_itp
            result['dif_inf_itp']= dif_inf_itp
            result['mse_inf'] = mse(result['label'], result['infer'])
            result['mse_itp'] = mse(result['label'], result['interp'])
            result['mse_inf_to_itp_ratio'] = result['mse_inf'] / \
                result['mse_itp']
        from dxpy.debug.utils import dbgmsg
        dbgmsg(result)
        return super()._processing(result)


"""
Config module make it easy for config intens projects.

Config module consists of following classes:
ConfigsViewer:
    Configs viewer class.

ConfigsMaker:
    

"""

# from .base import Configs
from .module_config import ModuleConfigs
from ._viewer import ConfigsView
from ._configurable import configurable



"""
This moudule is designed to be stateless (pure interface):
    No data members
    No Deamon process
Works
"""
import copy
import rx
import json
import rx
from dxpy.time.utils import now
from dxpy.time.timestamps import TaskStamp
from . import database
from .model import task as ts


def create(task) -> int:
    return database.create(task.to_json())


def create_graph(task_graph) -> 'list<int>':
    done = []
    for t in task_graph.nodes():
        t.id = None

    def all_dependency_added(task):
        return all([t in done for t in task_graph.dependencies(task)])

    def update_dependency(task):
        task.dependency = [t.id for t in task_graph.dependencies(task)]
        return task

    def add_to_done(task):
        done.append(task)
        return task

    def create_and_update_id(task):
        task.id = create(task)
        return task

    while len(done) < len(task_graph):
        (rx.Observable.from_(task_graph.nodes())
         .filter(lambda t: not t in done)
         .filter(all_dependency_added)
         .map(update_dependency)
         .map(create_and_update_id)
         .map(add_to_done)
         .subscribe())

    return [t.id for t in done]


def parse_json(s: 'json string'):
    return ts.Task.from_json(s)


def read(tid: int):
    if not isinstance(tid, int):
        raise TypeError("read only accept tid of int type: {!r}".format(tid))
    return parse_json(database.read(tid))


def read_all() -> 'Observable<TaskPy>':
    return (database.read_all()
            .map(parse_json))


def dependencies(task) -> 'Observable<TaskPy>':
    return rx.Observable.from_(task.dependency).map(read)


def update(task) -> None:
    database.update(task.to_json())
    return task


def mark_submit(task) -> None:
    return update(ts.submit(task))


def mark_start(task) -> None:
    if task.time_stamp.start is None:
        task.time_stamp = TaskStamp(create=task.time_stamp.create,
                                    start=now(),
                                    end=task.time_stamp.end)
    return update(ts.start(task))


def mark_complete(task) -> None:
    if task.time_stamp.end is None:
        task.time_stamp = TaskStamp(create=task.time_stamp.create,
                                    start=task.time_stamp.start,
                                    end=now())
    return update(ts.complete(task))


def delete(tid: int) -> None:
    database.delete(tid)



import os
from collections import UserDict
from dxpy.filesystem import Path
_default = {
    'path': str(Path(os.environ.get('PATH_DATABASE')) / 'taskdb.db'),
    'name': 'task',
    'names': 'tasks',
    'use_web_api': True,
    'echo': False,
    'default_state': 'BeforeSubmit',

    'host': '0.0.0.0',
    'ip': '127.0.0.1',
    'port': 23300,
    'debug': False,
    'version': 0.1,
    'base': '/'
}


class Config(UserDict):
    def __init__(self):
        super(__class__, self).__init__()
        self.data.update(_default)

    @property
    def path_sqllite(self):
        return 'sqlite:///' + self.data['path']

    def back_to_default(self):
        self.data.update(_default)


config = Config()



import tensorflow as tf

from dxpy.collections.dicts import combine_dicts

from ..graph import Graph, NodeKeys
import typing

from dxpy.learn.session import get_default_session
class SummaryItem:
    def __init__(self, tensor, stype=None):
        """
        Args:

        -   tensor: tensor to add to summary
        -   stype: `str`, indicating which summary type will be added.
                Currently support 'scalar', 'image'.
        """
        self.tensor = tensor
        if stype is None:
            if len(self.tensor.shape.as_list()) <= 1:
                stype = 'scalar'
            elif len(self.tensor.shape.as_list()) == 4:
                stype = 'image'
            else:
                raise TypeError(
                    "Can not auto infer summary type of {}.".format(tensor))
        self.stype = stype


class WithSummaryItems:
    def create_summary_items(self) -> typing.Dict[str, SummaryItem]:
        return dict()


class SummaryWriter(Graph):
    """
    Summary writer class.


    Usage:

        #create dataset, network

        dataset = ...

        network = ...

        summary = SummaryWriter('train', {'loss': network['loss']}, {'input': network['input'], 'label': network['label']} path='./summary/train/)

        #After create session
        sess = tf.Session()
        with sess.as_default():
            dataset.post_session_created()
            network.post_session_created()
            summary.post_session_created()

        #Train steps:
        with sess.as_default():
            for i in range(1000):
                network.train()
                if (i % 10) == 0:
                    summary.summary({'input': input_ndarray, 'label': label_ndarray})
            summary.flust()
    """

    def __init__(self, name='summary', tensors_to_summary=None, inputs=None, **config):
        """
        Inputs:
            tensors_to_summary: dict of name: SummaryItem
            inputs: tensors required to run summary (will used for feeds)
        """
        super().__init__(name, **config)
        if tensors_to_summary is None:
            tensors_to_summary = dict()
        if inputs is None:
            inputs = dict()

        self._tensors = combine_dicts(tensors_to_summary,
                                      self._default_summaries())
        for k in inputs:
            self.register_node(k, inputs[k])
        self._summary_ops = list()
        self.register_task('create_writer', self.__create_writer)
        self.register_task('flush', self.flush)
        self.register_main_task(self.summary)

        if self.param('add_prefix'):
            with tf.name_scope('summary'):
                self.__add_all_ops()
        else:
            self.__add_all_ops()

    @classmethod
    def _default_config(cls):
        return combine_dicts({'path': './summary',
                              'add_prefix': False,
                              'max_image': 3},
                             super()._default_config())

    @classmethod
    def _default_summaries(cls):
        return dict()

    def post_session_created(self):
        self.run('create_writer')

    def summary(self, feeds=None):
        from ..scalar import current_step
        if self.as_tensor() is None:
            return
        value = get_default_session().run(self.as_tensor(), self.get_feed_dict(feeds))
        self.nodes['summary_writer'].add_summary(value, current_step())

    def flush(self):
        self.nodes['summary_writer'].flush()

    def __add_all_ops(self):
        self.__add_tensors_to_summary()
        self.__add_main_op()

    def __add_main_op(self):
        if len(self._summary_ops) > 0:
            self.register_main_node(tf.summary.merge(self._summary_ops))

    def __create_writer(self, feeds):
        self.register_node('summary_writer',
                           tf.summary.FileWriter(self.param('path', feeds),
                                                 get_default_session().graph))

    def __add_tensors_to_summary(self):
        for n, v in self._tensors.items():
            if v is None:
                continue
            op = None
            if v.stype == 'scalar':
                op = tf.summary.scalar(n, v.tensor)
            elif v.stype == 'image':
                op = tf.summary.image(n, v.tensor)
            if op is not None:
                self._summary_ops.append(op)


class SummaryWriterForNet(SummaryWriter):
    """
    A quick summary writer which automaticlly add loss, evaluate (if exists),
    input, inference of net.
    """

    def __init__(self, name, net, tensors_to_summary=None, inputs=None, **config):
        automatic_summary_tensors = dict()
        automatic_keys = [NodeKeys.LOSS, NodeKeys.INFERENCE, NodeKeys.EVALUATE]
        for k in automatic_keys:
            if k in net.nodes:
                automatic_summary_tensors[k] = SummaryItem(net[k])
        if tensors_to_summary is None:
            tensors_to_summary = dict()
        super().__init__(name, combine_dicts(tensors_to_summary, automatic_summary_tensors),
                         inputs, **config)


from dxpy.configs import configurable
from dxpy.learn.config import config


class SummaryWriterV2(Graph):
    # TODO: Target features:
    # 1. summary_switch, within summary content, summaries will be added, otherwise ignored.
    # 2. get_summary_writer(), global access to summary writer, convinient for adding summries
    # 3. graph_name_filter: only response to summaries from given graph paths
    @configurable(config, with_name=True)
    def __init__(self, name='summary', tensors_to_summary=None, inputs=None,
                 *, valid_paths=None,
                 additional_paths=None,
                 **kw):
        super().__init__(name, valid_paths=valid_paths,
                         additional_paths=additional_paths, **kw)



import tensorflow as tf
from enum import Enum


class ActivationConfig:
    def __init__(self, name, pre_activation=False, post_activation=False):
        self.name = name
        self.pre_activation = pre_activation
        self.post_activation = post_activation


class ActivationConfigs(Enum):
    basic = ActivationConfig('relu', False, True)
    linear = ActivationConfig('none')
    none = ActivationConfig('none')
    res_celu = ActivationConfig('celu', True, False)
    incept = ActivationConfig('celu', True, False)
    pre = ActivationConfig('celu', True, False)    


def celu(x):
    """ like concatenated ReLU (http://arxiv.org/abs/1603.05201), but then with ELU """
    with tf.name_scope('celu'):
        x = tf.nn.elu(tf.concat([x, -x], axis=-1))
    return x


def apply(activation_config, input_tensor, position=None):
    if position is None:
        with tf.name_scope('activation'):
            return get_activation(activation_config.name)(input_tensor)
    elif position.lower() == 'pre' and activation_config.pre_activation:
        with tf.name_scope('pre_activation'):
            return get_activation(activation_config.name)(input_tensor)
    elif position.lower() == 'post' and activation_config.post_activation:
        with tf.name_scope('post_activation'):
            return get_activation(activation_config.name)(input_tensor)
    return input_tensor


def unified_config(name):
    if isinstance(name, ActivationConfig):
        return name
    for e in ActivationConfigs:
        if e == name or e.name == name:
            return e.value
    return None


def get_activation(name):
    name = name.lower()
    activations = {
        'relu': tf.nn.relu,
        'elu': tf.nn.elu,
        'celu': celu,
        'crelu': tf.nn.crelu,
        'none': lambda x: x}
    alias = {}
    for n in alias:
        if name in alias[n]:
            name = n
    if name in activations:
        return activations[name]
    raise ValueError("Unknown name of activation: {}.".format(name))



import tensorflow as tf
from dxpy.configs import configurable
from dxpy.learn.config import config



# @configurable(config, with_name=True)
# def start_parameter_server(cluster_spec_file='cluster.yml', job_name='ps', task_index=0, network_name=None, dataset_cluster_name=None, name='cluster/ps/task0'):
#     """
#     Args:
#         func: functions which returns dataset
#     """
#     from .cluster import get_cluster_spec
#     from .dataset import get_dist_dataset
#     cluster = tf.train.ClusterSpec(get_cluster_spec(cluster_spec_file))
#     server = tf.train.Server(cluster, job_name=job_name, task_index=task_index)
#     dataset = get_dist_dataset(name=dataset_cluster_name) 
#     network = get_dist_network(name=name, dataset=dataset)
#     server.join()

@configurable(config, with_name=True)
def get_dist_network(job_name='dataset', task_index=0, network_config=None, dataset=None, network_ps=None, name='cluster/ps/task0'):
    import tf.train.replica_device_setter
    if network_config is None:
        network_config = name
    from dxpy.learn.net.api import get_network
    with tf.device('/job:{}/task:{}'.format(job_name, task_index)):
    # with tf.device()
        network = get_network(name=network_config, dataset=dataset, network_ps=network_ps, reuse=True, scope=network_ps._scope)
        result = network()
    return network, result

@configurable(config, with_name=True)
def apply_dist_network(job_name='worker', task_index=0, network=None, dataset=None, name='cluster/ps/task0'):
    with tf.device('/job:{}/task:{}'.format(job_name, task_index)):
        return network(dataset)





class NotEqualLengthError(Exception):
    def __init__(self, names, lens):
        msg = "Lengths of lists or tuples are supposed to be equal, however get:\n"
        msg += "{:^32}{:^8}\n".format('Name or Array', 'Length')
        for n, l in zip(names, lens):
            msg += "{:^32}{:^8}\n".format(n, l)
        super(__class__, self).__init__(msg)

def assert_same_length(arrays, names=None):
    lens = [len(a) for a in arrays]
    if not all([l==lens[0] for l in lens]):
        if names is None:
            names = [repr(a) for a in arrays]
        raise NotEqualLengthError(names, lens)


import astra
import astra.creators as ac
import numpy as np
from dxpy.tensor.checks import assert_ndim
from dxpy.tensor.transform import maybe_unbatch
from ..config import config
from dxpy.configs import configurable


@configurable(config.get('projection'))
def projection2d(image, detector, *, method='cuda', projection_model='linear'):
    assert_ndim(image, 2, 'image')
    vol_geom = ac.create_vol_geom(image.shape)
    proj_geom = ac.create_proj_geom('parallel',
                                    detector.sensor_width,
                                    detector.nb_sensors,
                                    detector.views)
    proj_id = ac.create_projector(projection_model, proj_geom, vol_geom)
    sid, sinogram = ac.create_sino(image, proj_id, gpuIndex=0)
    astra.data2d.clear()
    astra.projector.clear()
    astra.algorithm.clear()
    return sinogram


class Projector2DParallel:
    @configurable(config.get('projection'))
    def __init__(self, detector, phantom_spec, *, method='cuda', projection_model='linear'):
        vol_geom = ac.create_vol_geom(phantom_spec.shape)
        proj_geom = ac.create_proj_geom('parallel',
                                        detector.sensor_width,
                                        detector.nb_sensors,
                                        detector.views)
        self.proj_id = ac.create_projector(
            projection_model, proj_geom, vol_geom)
        self.phan_id = astra.data2d.create('-vol', vol_geom)

    def project(self, image):
        assert_ndim(image, 2, 'image')
        astra.data2d.store(self.phan_id, image)
        sid, result = ac.create_sino(self.phan_id, self.proj_id)
        astra.data2d.delete(sid)
        return result

    def clear(self):
        astra.data2d.delete(self.phan_id)
        astra.projector.delete(self.proj_id)


def padding_pi2full(sinogram):
    """
    Padding sinogram of [0, np.pi) to full [0, 2*np.pi)
    """
    return np.concatenate([sinogram, np.flip(sinogram, 0)], axis=1)


import numpy as np
from . import tensor as te


def try_unbatch(images):
    if isinstance(images, np.ndarray) and images.ndim == 4:
        return te.unbatch(images)
    return images


def fix_dim(image):
    if image.ndim == 3 and image.shape[-1] == 1:
        return np.reshape(image, image.shape[:-1])
    return image



import tensorflow as tf
from ..model import Model
from ..graph import Graph, NodeKeys

from dxpy.configs import configurable
from dxpy.learn.config import config 
class Net(Model):
    """ Base class of nets.
    Net add some special tasks based on graph:
        1. train
        2. inference
        3. evaluate
        4. save/load

    Net will add these supports based on result of _kernel() calls.
    - NodeKeys.LOSS: scalar or List[scalar]
        If NodeKeys.Loss presented in self.nodes (automatically registed by model),
        a trainer will be added with name self.name / 'trainer'. Depending on type
        of loss, trainer will switch between single GPU and multi GPU version.
    - NodeKeys.INFERENCE: output of net, will run with self.inference(feeds)
        NOTE: feeds supports both registered name in self.nodes (str) and tf.Tensor
    - NodeKeys.EVALUATE: used for evaluate performace, will run in self.evaluate(feeds)
        NOTE: if no NodeKeys.EVALUATE in self.nodes, self.nodes[NodeKeys.LOSS] will be used. 
    """

    @configurable(config, with_name=True)
    def __init__(self, name, inputs=None, nb_evaluate_runs=32, **kw):
        super().__init__(name, inputs, nb_evaluate_runs=nb_evaluate_runs, **kw)

    @classmethod
    def _default_config(cls):
        from dxpy.collections.dicts import combine_dicts
        return combine_dicts({
            'add_trainer': True,
            'add_saver': True
        }, super()._default_config())

    def _tensors_need_summary(self):
        if NodeKeys.EVALUATE in self.nodes:
            if isinstance(self.nodes[NodeKeys.EVALUATE], tf.Tensor):
                return {NodeKeys.EVALUATE: self.nodes[NodeKeys.EVALUATE]}
            elif isinstance(self.nodes[NodeKeys.EVALUATE], dict):
                return self.nodes[NodeKeys.EVALUATE]
        return dict()

    def _post_kernel_post_outputs(self):
        super()._post_kernel_post_outputs()
        from ..train import Trainer, Saver
        if self.param('add_trainer'):
            from dxpy.debug.utils import dbgmsg
            dbgmsg(self.nodes[NodeKeys.LOSS])
            if NodeKeys.LOSS in self.nodes:
                self.register_node(NodeKeys.TRAINER,
                                   Trainer(self.name / 'trainer', self.nodes[NodeKeys.LOSS]))
        if self.param('add_saver'):
            self.register_node(NodeKeys.SAVER, Saver(self.name / 'saver'))

    def post_session_created(self):
        pass

    def summary_items(self):
        return dict()

    @property
    def session(self):
        from dxpy.learn.session import get_default_session
        return get_default_session()

    def train(self, feeds=None):
        self.nodes[NodeKeys.TRAINER].run('set_learning_rate')
        return self.nodes[NodeKeys.TRAINER](feeds)

    def inference(self, feeds=None):
        return self.session.run(self.tensor(NodeKeys.INFERENCE), self.get_feed_dict(feeds))

    def evaluate(self, feeds=None):
        return self.session.run(self.tensor(NodeKeys.EVALUATE), self.get_feed_dict(feeds))


    def save(self, feeds=None):
        return self.nodes[NodeKeys.SAVER].run('save', feeds)

    def load(self, feeds=None):
        return self.nodes[NodeKeys.SAVER].run('load', feeds)



"""
Fast implementation of layers, needs to be move to model implementation later.
"""

import tensorflow as tf




def upsampling2d(inputs, size, method=None, filters=None, name='upsampling'):
    if method is None:
        method = 'bilinear'
    input_shape = inputs.shape.as_list()
    h0, w0 = input_shape[1:3]
    h1 = int(h0 * size[0])
    w1 = int(w0 * size[1])
    with tf.name_scope(name):
        if method == 'nearest':
            h = tf.image.resize_nearest_neighbor(inputs, size=[h1, w1])
        elif method == 'bilinear':
            h = tf.image.resize_bilinear(inputs, size=[h1, w1])
        elif method == 'deconv':
            h = tf.layers.conv2d_transpose(
                inputs, filters, 3, strides=size, padding='same')
    return h


def enc_img(inputs, base_filters, data_format, reuse=None):
    conv_cfgs = {
        'activation': tf.nn.elu,
        'padding': 'same',
        'data_format': 'channel_first',
        'reuse': reuse
    }
    h = inputs
    f = base_filters
    with tf.name_scope('enc') as scope:
        h = tf.layers.conv2d(h, f, 3, **conv_cfgs, name=scope + 'conv1')
        h = tf.layers.conv2d(h, f, 3, **conv_cfgs, name=scope + 'conv2')
        h = tf.layers.conv2d(h, 2 * f, 2, 2, **conv_cfgs, name=scope + 'conv3')
        h = tf.layers.conv2d(h, 2 * f, 3, **conv_cfgs, name=scope + 'conv4')
        h = tf.layers.conv2d(h, 2 * f, 3, **conv_cfgs, name=scope + 'conv5')
        h = tf.layers.conv2d(h, 3 * f, 2, 2, **conv_cfgs, name=scope + 'conv6')
        h = tf.layers.conv2d(h, 3 * f, 3, **conv_cfgs, name=scope + 'conv7')
        h = tf.layers.conv2d(h, 3 * f, 3, **conv_cfgs, name=scope + 'conv8')
        h = tf.layers.conv2d(h, 4 * f, 2, 2, **conv_cfgs, name=scope + 'conv9')
        h = tf.layers.conv2d(h, 4 * f, 3, **conv_cfgs, name=scope + 'conv10')
        h = tf.layers.conv2d(h, 4 * f, 3, **conv_cfgs, name=scope + 'conv11')
        h = slim.flatten()
        shape_last = h.shape().to_list()
        dim_latent = int(np.prod(dim_latent[1:]))
        latent = tf.layers.dense(h, dim_latent, name=scope + 'dense')
    return latent


def dec_img(inputs, shape, base_filters, data_format, reuse=None):
    conv_cfgs = {
        'activation': tf.nn.elu,
        'padding': 'same',
        'data_format': data_format,
        'reuse': reuse
    }
    h = inputs
    f = base_filters
    with tf.name_scope('enc') as scope:
        h = tf.reshape(h, shape)
        h = tf.layers.conv2d(h, f, 3, **conv_cfgs, name=scope + 'conv1')
        h = tf.layers.conv2d(h, f, 3, **conv_cfgs, name=scope + 'conv2')
        h = tf.contrib.keras.layers.UpSampling2D(
            size=(2, 2), data_format=data_format)
        h = tf.layers.conv2d(h, 2 * f, 2, 2, **conv_cfgs, name=scope + 'conv3')
        h = tf.layers.conv2d(h, 2 * f, 3, **conv_cfgs, name=scope + 'conv4')
        h = tf.layers.conv2d(h, 2 * f, 3, **conv_cfgs, name=scope + 'conv5')
        h = tf.layers.conv2d(h, 3 * f, 2, 2, **conv_cfgs, name=scope + 'conv6')
        h = tf.layers.conv2d(h, 3 * f, 3, **conv_cfgs, name=scope + 'conv7')
        h = tf.layers.conv2d(h, 3 * f, 3, **conv_cfgs, name=scope + 'conv8')
        h = tf.layers.conv2d(h, 4 * f, 2, 2, **conv_cfgs, name=scope + 'conv9')
        h = tf.layers.conv2d(h, 4 * f, 3, **conv_cfgs, name=scope + 'conv10')
        h = tf.layers.conv2d(h, 4 * f, 3, **conv_cfgs, name=scope + 'conv11')
        h = slim.flatten()
        shape_last = h.shape().to_list()
        dim_latent = int(np.prod(dim_latent[1:]))
        latent = tf.layers.dense(h, dim_latent, name=scope + 'dense')
    return latent


def residual_unit(inputs, filters, kernel_size, strides=(1, 1), activation=tf.nn.elu, training=True, padding='same', data_format='channels_last', name='conv2d', nb_convs=2):
    h = inputs
    with tf.name_scope(name) as scope:
        for i in range(nb_convs):
            h = conv2d(h, filters, kernel_size, strides, activation,
                       training, padding, data_format, name='conv_%d' % i)


def sr_end(res, itp, ip_h, name='sr_end', is_res=True):
    """ Assuming shape(itp) == shape(ip_h)
    reps is center croped shape of itp/ip_h
    """
    with tf.name_scope(name):
        spo = res.shape.as_list()[1:3]
        spi = itp.shape.as_list()[1:3]
        cpx = (spi[0] - spo[0]) // 2
        cpy = (spi[1] - spo[1]) // 2
        crop_size = (cpx, cpy)
        itp_c = Cropping2D(crop_size)(itp)
        with tf.name_scope('output'):
            inf = add([res, itp_c])
        if is_res:
            with tf.name_scope('label_cropped'):
                ip_c = Cropping2D(crop_size)(ip_h)
            with tf.name_scope('res_out'):
                res_inf = sub(ip_c, inf)
            with tf.name_scope('res_itp'):
                res_itp = sub(ip_c, itp_c)
        else:
            res_inf = None
            res_itp = None
        return (inf, crop_size, res_inf, res_itp)


def inception_residual_block(input_tensor, filters, is_final_activ=False, is_bn=True, activation=tf.nn.crelu, training=True, reuse=None, res_scale=0.1, name='irb'):
    cc = {'reuse': reuse, 'padding': 'same'}
    cb = {'reuse': reuse, 'training': training, 'scale': False}
    with tf.name_scope(name):
        with tf.name_scope('nin'):
            if activation == tf.nn.crelu:
                input_tensor = tf.layers.conv2d(
                    input_tensor, 2 * filters, 1, **cc)
            else:
                input_tensor = tf.layers.conv2d(input_tensor, filters, 1, **cc)
        h1 = tf.layers.conv2d(input_tensor, filters, 1, **cc)
        if is_bn:
            h1 = tf.layers.batch_normalization(h1, **cb)
        h1 = activation(h1)
        h1 = tf.layers.conv2d(h1, filters, 3, **cc)

        h2 = tf.layers.conv2d(input_tensor, filters, 1, **cc)
        if is_bn:
            h2 = tf.layers.batch_normalization(h2, **cb)
        h2 = activation(h2)
        h2 = tf.layers.conv2d(h2, filters, 3, **cc)
        if is_bn:
            h2 = tf.layers.batch_normalization(h2, **cb)
        h2 = activation(h2)
        h2 = tf.layers.conv2d(h2, filters, 3, **cc)

        h = tf.concat([h1, h2], axis=-1)
        if is_bn:
            h = tf.layers.batch_normalization(h, **cb)
        h = activation(h)
        h = tf.layers.conv2d(h, filters, 1, **cc)
        if is_bn:
            h = tf.layers.batch_normalization(h, **cb)
        h = activation(h)

        h = h * res_scale
        h = h + input_tensor
        if is_final_activ:
            h = activation(h)
        return h


def align_by_crop(target_tensor, input_tensors, batch_size=None, name='align_crop'):
    target_shape = target_tensor.shape.as_list()
    if batch_size is not None:
        target_shape[0] = batch_size
    ops = []
    with tf.name_scope(name):
        for t in input_tensors:
            input_shape = t.shape.as_list()
            crop_h = (input_shape[1] - target_shape[1]) // 2
            crop_w = (input_shape[2] - target_shape[2]) // 2
            target_shape_now = list(target_shape)
            target_shape_now[-1] = input_shape[-1]
            ops.append(tf.slice(t, [0, crop_h, crop_w, 0], target_shape_now))
    return ops


def residual2(inputs, filters=64, activation=tf.nn.elu, scale=0.1, name='res_u', reuse=None):
    with tf.name_scope('residual_unit'):
        with tf.variable_scope(name, reuse=reuse):
            h = tf.layers.conv2d(
                inputs, filters, 3, padding='same', activation=activation, name='conv0')
            h = tf.layers.conv2d(h, filters, 1, padding='same',
                                 activation=activation, name='conv1')
            h = tf.layers.conv2d(h, filters, 3, padding='same', name='conv2')
        with tf.name_scope('shorcut'):
            out = scale * h + inputs
    return out


def nin(inputs, filters, activation=celu, name='nin', reuse=None, **kwargs):
    with tf.variable_scope(name, 'nin', reuse=reuse):
        h = conv2d(inputs, filters, 3)
        h = activation(h)
        h = conv2d(h, filters, 1)
        h = activation(h)
        h = conv2d(h, filters, 3)
    return h


def conv2d(inputs, filters, kernel_size=3, *, activation=None, normalization=None, training=None, name=None, reuse=None, padding='same', pre_activation=None, **kwargs):
    """ Warped conv2d

    """
    with tf.variable_scope(name, 'convolution2d', reuse=reuse):
        if pre_activation is not None:
            h = pre_activation(inputs)
        else:
            h = inputs
        h = tf.layers.conv2d(
            h, filters=filters, kernel_size=kernel_size, padding=padding, **kwargs)
        scale = not normalization is None
        if normalization == 'bn':
            h = tf.layers.batch_normalization(
                h, scale=scale, training=training)
        if activation is not None:
            h = activation(h)
    return h


def bn(*args, **kwargs):
    return tf.layers.batch_normalization(*args, **kwargs)


def stem(inputs, filters, reuse=None, name=None):
    """ Convolution start layer
    """
    with tf.variable_scope(name, 'stem', reuse=reuse):
        out = conv2d(inputs, filters)
    return out


def incept(inputs, filters, nb_path=3, activation=None, normalization=None, training=None, reuse=None, name=None):
    """ Basic incept unit.
    Inputs:
        inputs: input tensor
        filters: # of filters (maximum, summation of filters in seperated paths), # of channels of output tensor
        nb_path: # of paths
        activation: activation function
        normalization: normalization function name (currently 'bn' or None)
        training: if using batch normalization, training flag
        reuse: bool
        name: str
    """
    default_name = 'incept'
    fp = filters // nb_path
    with tf.variable_scope(name, default_name, reuse=reuse):
        with arg_scope([conv2d], padding='same'):
            with arg_scope([bn], training=training):
                # with tf.name_scope(name, default_name):
                h_paths = []
                for ipath in range(nb_path):
                    with tf.variable_scope('path_%d' % ipath):
                        h = inputs
                        kernel_sizes = [1] + [3] * ipath
                        for i, ks in enumerate(kernel_sizes):
                            if normalization == 'bn':
                                h = bn(h, name='bn_%d' % i)
                            if activation is not None:
                                h = activation(h)
                            h = conv2d(h, fp, kernel_size=ks,
                                       name='conv_%d' % i)
                        h_paths.append(h)
                # concate
                with tf.name_scope('concate'):
                    h = tf.concat(h_paths, axis=-1)
                with tf.variable_scope('path_merge'):
                    kernel_sizes = [1]
                    for i, ks in enumerate(kernel_sizes):
                        if normalization == 'bn':
                            h = bn(h, name='bn_%d' % i)
                        if activation is not None:
                            h = activation(h)
                        h = conv2d(h, filters, kernel_size=ks,
                                   name='conv_%d' % i)
    return h


def residual(inputs, basic_unit, basic_unit_args, nb_units=2, scale=None, reuse=None, name=None):
    """ Residual block.
        Automatically define filters, by adding a argument `filters` to basic_unit_args.
        basic_unit can be any layer which support args `filters` and `name`.
    """
    default_name = 'residual'
    filters = inputs.shape.as_list()[-1]
    basic_unit_args = dict(basic_unit_args)
    basic_unit_args.update({'filters': filters})
    if scale is None:
        scale = (1.0, 1.0)
    if not isinstance(scale, (list, tuple)):
        scale = (1.0, scale)
    h = inputs
    with tf.variable_scope(name, default_name, reuse=reuse):
        for i in range(nb_units):
            h = basic_unit(h, name='unit_%d' % i, **basic_unit_args)
        with tf.name_scope('add'):
            out = scale[0] * inputs + scale[1] * h
    return out


def repeat(inputs, layer, depth, layer_args, prefix, name=None):
    h = inputs
    with tf.variable_scope(name, 'repeat'):
        for i in range(depth):
            h = layer(h, name=prefix + '_%d' % i, **layer_args)
    return h


def super_resolution_infer(low_res, reps, size=(2, 2), reuse: bool=None, method=None, name: str=None):
    """Super resolution inference layer
    Inputs:
        low_res: low resolution tensor (NHW1)
        reps: representation of low resolution shape(NHWC)
        size: upsampling size
    Returns:
        `dict` of tensors:
            'inf': inference high resolution tensor
            'res': residual to interpolation tensor
            'itp': interpolation
    Yields:
        None
    """
    with tf.variable_scope(name, 'super_resolution_infer', reuse=reuse):
        interp = upsampling2d(low_res, size, method=method)
        residual = conv2d(reps, 1, kernel_size=1)
        with tf.name_scope('inference'):
            inference = interp + residual
    out = {
        'inf': inference,
        'res': residual,
        'itp': interp
    }
    return out


def crop(inputs, half_crop_size: list, offset: list=None, name='crop'):
    """Crop tensors.
    Inputs:
        half_crop_size: crop size of one direction
        offset: offset of begin. If `offset` is not None, no crop at begining after offset.
    Returns:
        croped tensor
    """
    with tf.name_scope(name):
        input_shape = inputs.shape.as_list()
        if offset is None:
            offset = half_crop_size
        target_shape = [s0 - o - c for s0, o,
                        c in zip(input_shape, offset, half_crop_size)]
        out = tf.slice(inputs, offset, target_shape)
    return out


def crop_many(inputs, crop_size=None, batch_size=None, name='crop_many'):
    """Crop tensors.
        If crop_size is None, tensors will be center cropped to same size as inputs[0].
    """
    if crop_size is None:
        keep_0 = True
        if len(inputs) < 2:
            raise ValueError(
                'Crop size must be specified when nb of inputs < 2.')
        else:
            crop_size = []
            sp0 = inputs[0].shape.as_list()
            sp1 = inputs[1].shape.as_list()
            csx = (sp1[1] - sp0[1]) // 2
            csy = (sp1[2] - sp0[2]) // 2
            crop_size = [csx, csy]
        target_shape = inputs[0].shape.as_list()
        if target_shape[0] is None:
            target_shape[0] = batch_size
    else:
        keep_0 = False
        if isinstance(crop_size, int):
            crop_size = [crop_size, crop_size]
        target_shape = inputs[0].shape.as_list()
        target_shape[1] -= 2 * crop_size[0]
        target_shape[2] -= 2 * crop_size[1]

    out = []
    for i, t in enumerate(inputs):
        if keep_0 and i == 0:
            out.append(t)
            continue
        t_c = tf.slice(t, [0, crop_size[0], crop_size[1], 0], target_shape)
        out.append(t_c)
    return out


def upsampling2d_many(inputs, size, method='nearest', reuse=None, name=None):
    out = []
    with tf.variable_scope(name, 'upsampling2d_many', reuse=reuse):
        for t in inputs:
            tu = upsampling2d(t, size=size, method=method)
            out.append(tu)
    return out


def log_sum_exp(x):
    """ numerically stable log_sum_exp implementation that prevents overflow """
    axis = len(x.get_shape()) - 1
    m = tf.reduce_max(x, axis)
    m2 = tf.reduce_max(x, axis, keep_dims=True)
    return m + tf.log(tf.reduce_sum(tf.exp(x - m2), axis))



import tensorflow as tf
from ...base import Net, NodeKeys
from dxpy.configs import configurable
from dxpy.learn.config import config


class SinNet(Net):
    @configurable(config, with_name=True)
    def __init__(self, name, inputs, nb_layers, nb_units, **kw):
        super().__init__(name=name, inputs=inputs, nb_layers=nb_layers,
                         nb_units=nb_units, add_trainer=True, add_saver=True, **kw)

    def _post_kernel_post_outputs(self):
        from dxpy.learn.train.trainer_2 import Trainer
        from dxpy.learn.train.saver import Saver
        if self.param('add_trainer'):
            if NodeKeys.LOSS in self.nodes:
                self.register_node(NodeKeys.TRAINER,
                                   Trainer(self.name / 'trainer', self.nodes[NodeKeys.LOSS]))
        if self.param('add_saver'):
            self.register_node(NodeKeys.SAVER, Saver(self.name / 'saver'))

    def _kernel(self, feeds):
        from dxpy.learn.model.losses import mean_square_error
        x = feeds[NodeKeys.INPUT]
        with tf.variable_scope('kernel'):
            h = x
            for i in range(self.param('nb_layers')):
                h = tf.layers.dense(h, self.param(
                    'nb_units'), activation=tf.nn.relu, name='dense_{}'.format(i))
            yp = tf.layers.dense(h, 1, name='infer')
        result = {NodeKeys.INFERENCE: yp}
        if NodeKeys.LABEL in feeds:
            y = feeds[NodeKeys.LABEL]
            loss = mean_square_error(y, yp)
            result[NodeKeys.LOSS] = loss
        return result



class UrlSpec:
    def __init__(self, protocol=None, username=None, password=None, resource=None, params=None):
        self.protocol = protocol
        self.username = username
        self.password = password
        self.resource = resource
        self.params = params

    def __eq__(self, url_spec):
        compares = [self.protocol == url_spec.protocol,
                    self.username == url_spec.username,
                    self.password == url_spec.password,
                    self.resource == url_spec.resource]
        return all(compares)

    @classmethod
    def parse(cls, s):
        p = opener.parse.parse_fs_url(s)
        return cls(p.protocol, p.username, p.password, p.resource, p.params)

    def serilization(self, format='JSON'):
        def to_json_dct():
            return {'protocol': self.protocol,
                    'username': self.username,
                    'password': self.password,
                    'resource': self.resource,
                    'params': self.params}
        if format.upper() == 'JSON':
            pass
        return None


QUOTED_SLASH = qut('/')
DOUBLE_QUOTED_SLASH = qut(qut('/'))


def is_quoted_url(s):
    return QUOTED_SLASH in s


def is_double_quoted_url(s):
    return DOUBLE_QUOTED_SLASH in s


def url_quote_path(path):
    return qut(path)


def url_unquote_path(url):
    return uqut(url)


def url_double_quoted_path(path):
    return qut(qut(path))


def url_double_unquoted_path(path):
    return uqut(uqut(path))


def unified_path_to_str(path):
    if path is None:
        raise NotValidPathError
    if isinstance(path, str):
        if is_quoted_url(path):
            path = url_unquote_path(path)
        return normpath(path)
    if isinstance(path, Path):
        return unified_path_to_str(str(path.path))
    if isinstance(path, pathlib.Path):
        return unified_path_to_str(str(path))


import tensorflow as tf
from dxpy.configs import configurable
from dxpy.learn.config import config



@configurable(config, with_name=True)
def start_parameter_server(cluster_spec_file='cluster.yml', job_name='ps', task_index=0, network_name=None, dataset_cluster_name=None, name='cluster/ps/task0'):
    """
    Args:
        func: functions which returns dataset
    """
    from .cluster import get_cluster_spec
    from .dataset import get_dist_dataset
    cluster = tf.train.ClusterSpec(get_cluster_spec(cluster_spec_file))
    server = tf.train.Server(cluster, job_name=job_name, task_index=task_index)
    dataset = get_dist_dataset(name=dataset_cluster_name) 
    network = get_dist_network(name=name, dataset=dataset)
    server.join()

@configurable(config, with_name=True)
def get_dist_network(job_name='ps', task_index=0, network_config=None, dataset=None, name='cluster/ps/task0'):
    if network_config is None:
        network_config = name
    from dxpy.learn.net.api import get_network
    with tf.device('/job:{}/task:{}'.format(job_name, task_index)):
        network = get_network(name=network_config, dataset=dataset)
    return network




import tensorflow as tf
from ...base import Net, NodeKeys
from ....model.cnn.residual import StackedResidualv2 
from dxpy.configs import configurable
from dxpy.learn.config import config

class VeryDeepDenoiseNetwork(Net):
    @configurable(config, with_name=True)
    def __init__(self, name='VDDN', inputs=None, filters=32, **kw):
        super().__init__(name, inputs, filters=stem_filters, **kw)

    def summary_items(self):
        pass

    def _make_infer(self, input_):
        ip = input_
        stem = tf.layers.conv2d(ip, self.param('filters'), 3, padding='same', name='stem')
        x = StackedResidualv2(self.name/'kernel', stem)
        resi = tf.layers.conv2d(x, 1, 3, padding='same', name='residual')
        infer = resi + ip

    def _make_loss(self, label, target):
        from dxpy.learn.model.losses import mean_square_error
        return mean_square_error(label, target)

    def _kernel(self, feeds):
        from dxpy.learn.model.image import align_crop
        from dxpy.learn.utils.general import device_name

        ip = feeds[NodeKeys.INPUT]
        infer = self._make_infer(ip)
        result = {NodeKeys.INFERENCE: infer}
        if NodeKeys.LABEL in feeds:
            loss = self._make_loss(feeds[NodeKeys.LABEL, infer]) 
            result.update(NodeKeys.LOSS: loss)
        return result
        




from dxpy.collections import TreeDict
from ruamel.yaml import YAML
yaml = YAML()

config = TreeDict()

from dxpy.configs import ConfigsView


# TODO: Add support for global configs
# def load_default():
#     from dxpy.code import path_of_global_defaults
#     path_global_config = Path(path_of_global_defaults / 'xlean.yml').abs
#     with open('path')
#     global_config = yaml.load(path)




import tensorflow as tf
from ..config import config
from dxpy.configs import configurable


def _split_feeds(feeds, nb_gpu):
    from .tensor import MultiGPUSplitor, PlaceHolder
    mgs = MultiGPUSplitor(nb_gpu=nb_gpu)
    feeds_gpus = mgs(feeds)
    first_part = list(feeds_gpus.values())[0]
    with tf.name_scope('cpu_placeholders'):
        cpu_placeholders = {k: PlaceHolder(
            first_part[k]).as_tensor() for k in first_part}
    return feeds_gpus, mgs.part_names(), cpu_placeholders


def _apply_on_gpus(model, feeds_gpu, part_names):
    result_gpus = {}
    for i, k in enumerate(part_names):
        with tf.device(device_name('gpu', i)):
            result_gpus[k] = model(feeds_gpus[k])
    return result_gpus


def _merge(result_gpus, part_names):
    with tf.device(device_name('cpu')):
        with tf.name_scope('merge'):
            with tf.name_scope('inference'):
                infer = tf.concat([result_gpus[k][NodeKeys.INFERENCE]
                                   for k in part_names], axis=0)
            losses = [result_gpus[k][NodeKeys.LOSS]
                      for k in part_names]
            with tf.name_scope('total_loss'):
                loss = tf.add_n(losses)
    return {NodeKeys.INFERENCE: infer,
            NodeKeys.LOSS: losses,
            NodeKeys.EVALUATE: loss}


def apply_multi_gpu(feeds, model_func, nb_gpu=None):
    if nb_gpu is None:
        nb_gpu = config['nb_gpu']
    feeds_gpus, part_names, cpu_placeholders = _split_feeds(feeds, nb_gpu)
    with tf.device(device_name('cpu')):
        model = model_func(cpu_placeholders)
        model()
    result_gpus = _apply_on_gpus(model, feeds_gpu, part_names)
    return _merge(result_gpus, part_names)


class MultiGPUSplitor(Model):
    """
    A Helper class to create
    """
    @configurable(config.get('gpu'))
    def __init__(self, name='gpu_splitor', nb_gpu=2):
        super().__init__(name, nb_gpu=nb_gpu)

    @classmethod
    def _default_config(cls):
        result = dict()
        result.update(super()._default_config())
        result.update({
            'lazy_create': True,
            'register_inputs': False,
            'register_outputs': False,
            'format': 'split'
        })
        return result

    def _kernel(self, feeds):
        from dxpy.collections.dicts import swap_dict_hierarchy
        result = {k: self._slice_tensor_for_gpu(feeds[k]) for k in feeds}
        if self.param('format') == 'name':
            return result
        else:
            return swap_dict_hierarchy(result)

    def _get_shape_split(self, shape):
        if isinstance(shape, tf.TensorShape):
            shape_gpu = shape.as_list()
        else:
            shape_gpu = list(shape)
        if shape_gpu[0] is None:
            raise TypeError("Batch size can not be None for MultiGPUSplitor.")
        shape_gpu[0] = shape_gpu[0] // self.param('nb_gpu')
        return shape_gpu

    def _slice_tensor_for_gpu(self, tensor_input):
        result = dict()
        shape_gpu = self._get_shape_split(tensor_input.shape)
        with tf.name_scope(self._get_tensor_name(tensor_input.name)):
            for id_slice in range(self.param('nb_gpu')):
                with tf.device(device_name('gpu', id_slice)):
                    slice_start = [id_slice * shape_gpu[0]] + \
                        [0] * (len(shape_gpu) - 1)
                    name = 'part_{}'.format(id_slice)
                    result[name] = tf.slice(tensor_input, slice_start,
                                            shape_gpu, name=name)
        if len(result) == 0:
            return tensor_input
        return result

    def part_names(self):
        return ['part_{}'.format(id_slice) for id_slice in range(self.param('nb_gpu'))]

    def _get_tensor_name(self, name):
        prefix, idt = name.split(':')
        idt = int(idt)
        if idt == 0:
            return prefix
        else:
            return '{}_{}'.format(prefix, idt)



from .base import create, read, read_all, update, delete, Database
from .cli import cli
from .web import add_api
from ..config import config
from ..service.base import check_json
from ..exceptions import *



import tensorflow as tf
from fs.osfs import OSFS


def dummy_image(shape=None, batch_size=32, nb_channel=1, name='dummy_image'):
    if shape is None:
        shape = [28, 28]
    return tf.placeholder(tf.float32, [batch_size] + shape + [nb_channel], name=name)


def write_graph():
    from .general import pre_work
    from ..train.summary import SummaryWriter
    from ..session import Session
    with OSFS('/tmp') as fs:
        dir_name = 'tf_tests'
        if fs.exists(dir_name):
            d = fs.opendir(dir_name)
        else:
            d = fs.makedir(dir_name)
        if d.exists('summary/train'):
            d.removetree('summary/train')
        summary = SummaryWriter(name='train',
                                path=d.getsyspath('summary/train'))
        session = Session()
        with session.as_default():
            summary.post_session_created()



from typing import Dict, List, TypeVar
import arrow

import tensorflow as tf

from dxpy.collections.dicts import combine_dicts
from dxpy.configs import configurable
from dxpy.core.path import Path
from dxpy.learn.config import config
from dxpy.learn.session import get_default_session
import arrow
import time
from ..graph import Graph, NodeKeys


def get_summary_type(tensor, summary_type):
    from dxpy.learn.model.tensor import shape_as_list
    if summary_type is None:
        shape = shape_as_list(tensor)
        if len(shape) == 0 or (len(shape) == 1 and shape[0] == 1):
            summary_type = 'scalar'
        elif len(shape) == 4:
            summary_type = 'image'
        else:
            raise TypeError(
                "Can not auto infer summary type of {}.".format(tensor))
    return summary_type


def get_merge_method(tensor, merge_method):
    if merge_method is None:
        if get_summary_type(tensor, None) == 'scalar':
            return 'mean'
        elif get_summary_type(tensor, None) == 'image':
            return 'concat'
    else:
        return merge_method


def create_summary_op(name, tensor, summary_type=None):
    summary_type = get_summary_type(tensor, summary_type)
    if summary_type == 'scalar':
        return tf.summary.scalar(name, tensor)
    elif summary_type == 'image':
        return tf.summary.image(name, tensor)
    raise ValueError("Unknown summary_type {}.".format(summary_type))


class SummaryKeys:
    MERGED = 'merged'


class SummaryWriter(Graph):
    """
    Summary writer class.
    Usage:

        #create dataset, network
        dataset = ...
        network = ...
        summary = SummaryWriter('train', {'loss': network['loss']}, {'input': network['input'], 'label': network['label']}, path='./summary/train/)

        #After create session
        sess = tf.Session()
        with sess.as_default():
            dataset.post_session_created()
            network.post_session_created()
            summary.post_session_created()

        #Train steps:
        with sess.as_default():
            for i in range(1000):
                network.train()
                if (i % 10) == 0:
                    summary.summary({'input': input_ndarray, 'label': label_ndarray})
            summary.flust()
    """
    @configurable(config, with_name=True)
    def __init__(self, name: str='summary',
                 tensors: Dict[str, tf.Tensor]=None,
                 additional_inputs: Dict[str, tf.Tensor]=None,
                 nb_runs: Dict[str, int]=None,
                 merge_method: Dict[str, str]=None,
                 *,
                 path: str='./summary/train',
                 nb_max_image: int=3,
                 interval: int=600,
                 with_prefix: bool=False,
                 **kw):
        """
        Args:
            tensors: dict of name: SummaryItem
            inputs: tensors required to run summary (will used for feeds)
        """
        super().__init__(name, path=path,
                         nb_max_image=nb_max_image, with_prefix=with_prefix,
                         interval=interval, **kw)
        if tensors is None:
            tensors = dict()
        if additional_inputs is None:
            additional_inputs = dict()
        if nb_runs is None:
            nb_runs = dict()
        if merge_method is None:
            merge_method = dict()
        self.nb_runs = nb_runs
        self.merge_method = merge_method
        self.multi_runs = dict()
        self.nb_max_runs = 0
        self.inputs = dict()
        with tf.name_scope(self.basename):
            tensors = self._processing(tensors)
            summary_ops = self._create_summary_ops(tensors)
            with tf.name_scope('merge_summary_ops'):
                self.register_node(SummaryKeys.MERGED,
                                   tf.summary.merge(summary_ops))
        self.register_task('create_writer', self._create_writer)
        self.register_task('flush', self.flush)
        self._pre_summ = arrow.now()

    def _processing(self, tensors: Dict[str, tf.Tensor]) -> Dict[str, tf.Tensor]:
        result = dict()
        for k in tensors:
            if k in self.nb_runs:
                self.inputs[k] = tensors[k]
                with tf.name_scope('multi_runs'):
                    self.multi_runs[k] = [tf.placeholder(
                        tensors[k].dtype, tensors[k].shape, name=k) for i in range(self.nb_runs[k])]
                    if self.nb_runs[k] > self.nb_max_runs:
                        self.nb_max_runs = self.nb_runs[k]
                    mm = get_merge_method(tensors[k], self.merge_method.get(k))
                    if mm == 'mean':
                        reduce_mean = tf.add_n(
                            self.multi_runs[k]) / len(self.multi_runs[k])
                        result[k] = reduce_mean
                    elif mm == 'concat':
                        concat = tf.concat(self.multi_runs[k], axis=0)
                        result[k] = concat
            else:
                result[k] = tensors[k]
        return result

    def _create_summary_ops(self, tensors):
        result = list()
        for k in tensors:
            summ_op = create_summary_op(k, tensors[k])
            result.append(summ_op)
        return result

    def post_session_created(self):
        self.run('create_writer')

    def _get_multiple_run_feeds(self, feeds=None):
        result = dict()
        for i in range(self.nb_max_runs):
            current_result = get_default_session().run(
                self.inputs, self.get_feed_dict(feeds))
            for k in self.inputs:
                if k in self.nb_runs and self.nb_runs[k] > i:
                    result[self.multi_runs[k][i]] = current_result[k]
        return result

    def summary(self, feeds=None):
        from ..scalar import current_step
        mrf = self._get_multiple_run_feeds(feeds)
        feed_dict = self.get_feed_dict(feeds)
        feed_dict.update(mrf)
        value = get_default_session().run(
            self.nodes[SummaryKeys.MERGED], feed_dict=feed_dict)
        self.nodes['summary_writer'].add_summary(value, current_step())

    def auto_summary(self, feeds=None):
        time_now = arrow.now()
        dtime = (time_now - self._pre_summ).seconds
        if dtime > self.param('interval'):
            self.summary(feeds)
            print('Add Summary at {}'.format(arrow.now()))
            self._pre_summ = time_now
        else:
            time.sleep(self.param('interval') - dtime)

    def flush(self):
        self.nodes['summary_writer'].flush()

    def _create_writer(self, feeds):
        self.register_node('summary_writer',
                           tf.summary.FileWriter(self.param('path', feeds),
                                                 get_default_session().graph))



import tensorflow as tf
from ..graph import Graph, NodeKeys


class Model(Graph):
    """
    Graphs which do not support sofiscated tasks, just acts as an function.
    NOTE: Model is using Tensorflow resue scheme. Thus by using its name.
    Sharing model is convinient within one model, but is hard (which needs supports for child model) between models.
    A Model object is mainly used for two purpose:
        1. Share weights / reuse model for different inputs
        2. Access global configs as it inherent Graph
        3. Use its 'lazy_create' config to for some case of defining graph,
        which is useful for simiplified managing tensorflow name scopes.

    Configs:
        reuse: bool
        tensorflow_variable_scope: str=None, *NOT IMPLEMENTED YET*.
        lazy_create: bool=False, if is True, will not create graph on __init__.

    Model extends Graphs interface with two properities:
        inputs
        outputs

    They are implemented based on nodes.
    Model will generate output tenors in the following two setuations:
        1. when model is created;
        2. when apply (or __call__) method is called.
    You may merge input/output tensors by using 'lazy_create' option, in that case,
    model will not create placeholder inputs, and will generate output tensors in
    the first time it was called.

    Examples:
        1. Model used for only once:
            z = UserModel('model1', {'input1': x, 'input2': y})()
        2. Reuse model
            m = UserModel('model1', {'input1': PlaceHolder(x.shape), 'input2': y})
            z1 = m({input1': x1})
            z2 = m({input1': x2})
            y is shared for inputs of z1 and z2.
    """

    def __init__(self, name, inputs=None, child_models=None, *, lazy_create=None, scope=None, reuse=None, register_inputs=None, register_outputs=None, simple_output=None, **config):
        """
        Inputs:
            name: Path, name of model.
            inputs: tf.Tensor | dict of str: tf.Tensor|dict|Graph
                case tf.Tensor:
                    add NodeKeys.INPUT: tf.Tensor
                case name: tf.Tensor:
                    add name: tf.Tensor to _inputs
                case name: dict
                    add name: tf.placeholder with dict['shape'], dict['type]
                case name: dict
                    add name: Graph.as_tensor()
            child_models: dict of str: Model|function

        Args:
            simple_output: bool, if True, return result[NodeKeys.Output] directly instead of {NodeKeys.Output: tensor}
        """
        super().__init__(name, lazy_create=lazy_create, reuse=reuse, register_inputs=register_inputs,
                         register_outputs=register_outputs, simple_output=simple_output, **config)
        self._created = False
        self._scope = scope
        self._inputs = self._unified_inputs(inputs)
        self._child_models = child_models
        if not self.param('lazy_create'):
            self._construct()
        self.register_main_task(self.apply)

    @classmethod
    def _default_config(cls):
        from dxpy.collections.dicts import combine_dicts
        cfg = {
            'lazy_create': False,
            'reuse': None,
            'register_inputs': True,
            'register_outputs': True,
            'simple_output': True,
            'no_new_output_for_none_feeds': True,
            'register_output_with_prefix': True,
        }
        return combine_dicts(cfg, super()._default_config())

    @classmethod
    def _default_inputs(cls):
        return dict()

    def _post_create(self):
        pass

    def _pre_kernel_pre_inputs(self):
        pass

    def _pre_kernel_post_inputs(self):
        """
        Function hook before kernel of after register inputs during create.
        Useful to create child models.
        """
        pass

    def _post_kernel_pre_outputs(self):
        pass

    def _post_kernel_post_outputs(self):
        pass

    def _post_create(self):
        pass

    def _kernel(self, feeds):
        return feeds

    def apply(self, feeds=None):
        if not self._created:
            self._inputs = self._unified_inputs(feeds)
            self._construct()
            result = self._outputs
        else:
            if feeds is None:
                result = self._outputs
            else:
                with tf.variable_scope(self._variable_scope, reuse=True):
                    inputs = self._unified_inputs(feeds)
                    result = self._unified_outputs(
                        self._kernel(inputs))
        return self._simplified_output(result)

    def register_child_model(self, name, model):
        self._child_models[name] = model
        self.register_node('child_model/{}'.format(name), model)
        return model

    def get_model(self, name, create_func=None):
        if self._child_models.get(name) is None:
            if create_func is not None:
                self.register_child_model(name, create_func(self.name / name))
            else:
                fmt = "Can not found child model {} in {}."
                raise KeyError(fmt.format(name, str(self.name)))
        return self._child_models.get(name)

    def __create_non_tensor_inputs(self):
        from .tensor import PlaceHolder
        with tf.name_scope('inputs'):
            inputs = self._unified_inputs()
            for n in inputs:
                if isinstance(inputs[n], tf.Tensor):
                    self.register_node(n, inputs[n])
                else:
                    if isinstance(inputs[n], PlaceHolder):
                        shape = inputs[n].shape
                        dtype = inputs[n].dtype
                    else:
                        shape = inputs[n]['shape']
                        dtype = tf.float32
                    inputs[n] = self.create_placeholder_node(
                        tf.float32, inputs[n]['shape'], n)
                    self._inputs[n] = inputs[n]

    def __create_non_model_child_models(self):
        if self._child_models is None:
            self._child_models = dict()
            return
        if not isinstance(self._child_models, dict):
            if isinstance(self._child_models, Model):
                self._child_models = {NodeKeys.CHILD_MODEL: self._child_models}
            else:
                self._child_models = {
                    NodeKeys.CHILD_MODEL:  self._child_models()}
            return
        for n in self._child_models:
            if not isinstance(self._child_models[n], Model):
                self._child_models[n] = self._child_models[n]()

    def __register_inputs(self):
        if self.param('register_inputs'):
            for n in self._inputs:
                self.register_node('inputs/{}'.format(n), self._inputs[n])

    def __register_outputs(self):
        if self.param('register_outputs'):
            for n in self._outputs:
                if self.param('register_output_with_prefix'):
                    output_key = 'outputs/{}'.format(n)
                else:
                    output_key = n
                self.register_node(output_key, self._outputs[n])
                left_out_keys = [NodeKeys.MAIN,
                                 NodeKeys.LOSS, NodeKeys.INFERENCE]
                if n in left_out_keys:
                    self.register_node(n, self._outputs[n])
    def __create_kernel_in_scope(self):
        self._pre_kernel_pre_inputs()
        self.__create_non_tensor_inputs()
        self.__register_inputs()
        # self.__create_non_model_child_models()
        self._pre_kernel_post_inputs()
        self._outputs = self._unified_outputs(self._kernel(self._inputs))
        self._post_kernel_pre_outputs
        self.__register_outputs()
        self._post_kernel_post_outputs()       
    def __create(self):
        if self._created:
            raise ValueError(
                "Model {} is created. Can not recreate.".format(self.name))
        if self.param('reuse', raise_key_error=False):
            if self._scope is None:
                raise TypeError("Need scope for explicit reuse models")
            with tf.variable_scope(self._variable_scope, reuse=True):
                self.__create_kernel_in_scope()
        else:
            with tf.variable_scope(self._variable_scope, reuse=False) as scope:
            # with tf.variable_scope(self._variable_scope) as scope:
                self._scope = scope
                self.__create_kernel_in_scope()
        self._created = True

    def _construct(self):
        self._post_create()
        self.__create()
        self._post_create()

    @property
    def _variable_scope(self):
        if self._scope is None:
            return self.basename
        else:
            return self._scope

    @classmethod
    def _tensor_dict_standardization(cls, tensors=None, default_key=None):        
        from .tensor import PlaceHolder
        if tensors is None:
            return dict()

        if isinstance(tensors, (tf.Tensor, PlaceHolder)):
            return {default_key: tensors}
        result = dict()
        for n in tensors:
            if tensors[n] is not None:
                result[n] = tensors[n]
        return result

    def _unified_inputs(self, inputs=None):
        result = self._default_inputs()
        if hasattr(self, '_inputs'):
            result.update(self._inputs)
        result.update(self._tensor_dict_standardization(inputs,
                                                        NodeKeys.INPUT))
        return result

    def _unified_outputs(self, outputs):
        result = dict()
        result.update(self._tensor_dict_standardization(outputs,
                                                        NodeKeys.MAIN))
        return result

    def _simplified_output(self, results):
        if self.param('simple_output') and len(results) == 1 and list(results.keys())[0] == NodeKeys.MAIN:
            return results[NodeKeys.MAIN]
        return results


class ModelPipe(Model):
    def __init__(self, input_tensor, name, **config):
        super().__init__(name, {NodeKeys.INPUT: input_tensor}, **config)


models = dict()


def register_model(key, model):
    if not isinstance(model, Model):
        raise TypeError("Only Model is supported, got {}.".format(type(model)))
    if key in models:
        raise ValueError("Key {} already exists.".format(key))
    models[key] = model


def load_model(key):
    return models.get(key)



import numpy as np
from typing import List, TypeVar


def random_crop_offset(input_shape, target_shape, *, batched=False):
    max_offset = [s - t for s, t in zip(input_shape, target_shape)]
    if any(map(lambda x: x < 0, max_offset)):
        raise ValueError("Invalid input_shape {} or target_shape {}.".format(
            input_shape, target_shape))
    if not batched:
        offset = []
        for s in max_offset:
            if s == 0:
                offset.append(0)
            else:
                offset.append(np.random.randint(0, s))
        return offset
    else:
        offsets = []
        if max_offset[0] > 0:
            raise ValueError(
                "Random crop offset input_shape[0] and target_shape[0]")


def unbatch(tensor):
    if isinstance(tensor, np.ndarray):
        return list(map(lambda x: x[0, ...], np.split(tensor, tensor.shape[0])))
    raise TypeError("numpy.ndarray is required, got {}.".format(type(tensor)))


def maybe_unbatch(tensors: TypeVar('T', np.ndarray, List[np.ndarray])) -> List[np.ndarray]:
    if isinstance(tensors, (list, tuple)):
        return tensors
    else:
        return unbatch(tensors)


def padding(tensor: np.ndarray, target_shape: List[int], offset: TypeVar('T', int, List[int])=0, method: TypeVar('T', str, List[str])=None, *, padding_order: List[int]=None, constant_value: float=None):
    """
    Supported method:
        period: periodically padding tensor, assuming period equals to tensor shape.
        mirror: mirror reflect
        constant: pad with constant value
    """
    for i, v in enumerate(target_shape):
        if v is None:
            target_shape[i] = tensor.shape[i]
    if isinstance(offset, int):
        offset = [offset] * tensor.ndim
    if isinstance(method, str):
        method = [method] * tensor.ndim
    if padding_order is None:
        padding_order = list(range(tensor.ndim))

    def padding_one_block(source, target, idx, ids, method, mirror_axis=None):
        if method is None:
            target[idx] = source[ids]
        elif method.lower() == 'constant':
            target[idx] = constant_value
        elif method.lower() == 'period':
            target[idx] = source[ids]
        elif method.lower() == 'mirror':
            ids[mirror_axis] = slice(ids[mirror_axis].stop-1, None, -1)
            target[idx] = source[ids]

    def index(tensor, target_size, axis, offset):
        ids = list()
        idt = list()
        for i in range(tensor.ndim):
            if i == axis:
                max_idx = target_size
                slice_min = min([max([offset, 0]), max_idx])

                slice_max = min(
                [max([offset + tensor.shape[i], 0]), max_idx])
            else:
                slice_min = 0
                slice_max = tensor.shape[i]
            idt.append(slice(slice_min, slice_max))
            ids.append(slice(0, slice_max - slice_min))
        return idt, ids

    def current_method(method, is_mirror):
        if method == 'mirror':
            if is_mirror:
                c_method = 'mirror'
                is_mirror = False
            else:
                c_method = 'period'
                is_mirror = True
        else:
            c_method = method
        return c_method, is_mirror

    def padding_one_dim(tensor, axis, target_size, offset, method):
        offset_start = offset
        offset_now = offset_start
        next_shape = list(tensor.shape)
        next_shape[axis] = target_size
        result = np.zeros(next_shape)
        is_mirror = False
        while offset_now < target_size:
            idt, ids = index(tensor, target_size, i, offset_now)
            c_method, is_mirror = current_method(method, is_mirror)
            padding_one_block(tensor, result, idt, ids, c_method, axis)
            offset_now += tensor.shape[axis]
        is_mirror = True
        offset_now = offset_start - tensor.shape[axis]
        while offset_now > - tensor.shape[axis]:
            idt, ids = index(tensor, target_size, i, offset_now)
            c_method, is_mirror = current_method(method, is_mirror)
            padding_one_block(tensor, result, idt, ids, c_method, axis)
            offset_now -= tensor.shape[axis]
        return result
    result = np.array(tensor)
    for i in padding_order:
        if method[i] is None:
            continue
        result = padding_one_dim(
            result, i, target_shape[i], offset[i], method[i])
    return result



from pprint import pprint

import numpy as np
import tensorflow as tf
from dxpy.learn.dataset.zoo.sraps import AnalyticalPhantomSinogramDatasetForSuperResolution
from dxpy.learn.train.summary import SummaryWriter
from dxpy.learn.utils.general import pre_work
from dxpy.learn.model.cnn.super_resolution import SuperResolutionMultiScale, BuildingBlocks
from tqdm import tqdm
from dxpy.learn.utils.general import load_yaml_config

def dataset():
    load_yaml_config('dxln.yml')
    dataset = AnalyticalPhantomSinogramDatasetForSuperResolution()
    inputs = {}
    for i in range(dataset.param('nb_down_sample') + 1):
        inputs.update({'input/image{}x'.format(2**i): dataset['input/image{}x'.format(2**i)]})
        if external_configs['use_noise_label']:
            inputs.update({'label/image{}x'.format(2**i): dataset['input/image{}x'.format(2**i)]})
        else:
            inputs.update({'label/image{}x'.format(2**i): dataset['label/image{}x'.format(2**i)]})
            if dataset.param('image_type') == 'image':
                inputs['label/image1x'] = dataset['label/phantom']
    return inputs


import numpy as np
import tensorflow as tf
from ..graph import Graph, NodeKeys

from dxpy.configs import configurable
from ..config import get_config, config


class DatasetClassic(Graph):
    """Base class of database
    Database add some special tasks based on graph:
        1. single
        2. batch
    These methods returns a dict of Tensors:


    Parameters:
        batch_size: int, batch size of samples
        fields:
            key, shape
    """

    def __init__(self, name, **config):
        super(__class__, self).__init__(name, **config)
        with tf.name_scope(self.basename):
            for f in self.param('fields'):
                field_config = self.param('fields')[f]
                self.create_placeholder_node(
                    field_config['dtype'], self.batched_shape(f), f)
        self.register_task('single', self.single)
        self.register_task('batch', self.batch)

    def _load_sample(self, feeds=None):
        raise NotImplementedError

    def _load_dummpy(self, feeds=None):
        raise NotImplementedError

    def get_feed_dict(self, task=None):
        result = dict()
        batched_data = self.batch()
        for f in self.param('fields'):
            result[self.nodes[f]] = batched_data[f]
        return result

    def single(self, feeds=None):
        try:
            return self._load_sample(feeds)
        except StopIteration:
            return self._load_dummpy(feeds)

    def batch(self, feeds=None):
        result = dict()
        for f in self.param('fields', feeds):
            result[f] = np.zeros(self.batched_shape(f))
        for i in range(self.param('batch_size', feeds)):
            next_sample = self.single(feeds)
            for k in next_sample:
                result[k][i, ...] = next_sample[k]
        return result

    def batched_shape(self, field_name):
        return list([self.param('batch_size')] +
                    list(self.param('fields')[field_name]['shape']))


class DatasetBase(Graph):
    def __init__(self, name, **config):
        super().__init__(name, **config)
        self._pre_processing()
        with tf.name_scope(self.basename):
            self.dataset = self._processing()
            self.__register_dataset()
            self._post_processing()

    def _pre_processing(self):
        pass

    def _post_processing(self):
        # for k in self.as_tensor():
        pass

    def post_session_created(self):
        pass

    def _processing(self):
        raise NotImplementedError

    def __register_dataset(self):
        iterator = self.dataset.make_one_shot_iterator()
        next_element = iterator.get_next()
        self.register_main_node(next_element)
        for k in next_element:
            self.register_node(k, next_element[k])


class DatasetTFRecords(DatasetBase):
    def __init__(self, name, **config):
        super().__init__(name, **config)

    def _load_tfrecord_files(self):
        from dxpy.batch import FilesFilter
        from fs.osfs import OSFS
        return tf.data.TFRecordDataset(self.param('files'))

    def _processing(self):
        return self._load_tfrecord_files()


class DatasetBasev2(Graph):
    @configurable(get_config(), with_name=True)
    def __init__(self, name, batch_size=8, **config):
        super().__init__(name, batch_size=batch_size, **config)
        self._pre_processing()
        with tf.name_scope(self.basename):
            self.dataset = self._processing()
            self.__register_dataset()
            self._post_processing()

    def _pre_processing(self):
        pass

    def _post_processing(self):
        from dxpy.learn.utils.tensor import ensure_shape
        for k in self.as_tensor():
            self.nodes[k] = ensure_shape(
                self.nodes[k], batch_size=self.param('batch_size'))

    def post_session_created(self):
        pass

    def _processing(self):
        raise NotImplementedError

    def __register_dataset(self):
        iterator = self.dataset.make_one_shot_iterator()
        next_element = iterator.get_next()
        self.register_main_node(next_element)
        for k in next_element:
            self.register_node(k, next_element[k])


class DatasetFromTFDataset(DatasetBasev2):
    @configurable(get_config(), with_name=True)
    def __init__(self, name, tf_dataset, **config):
        self.tf_dataset = tf_dataset
        super().__init__(name, **config)

    def _processing(self):
        return self.tf_dataset

class DatasetFromGenerator(DatasetBasev2):
    @configurable(config, with_name=True)
    def __init__(self, name, generator, output_types, output_shapes, **config):
        self.generator = generator
        super().__init__(name, output_types=output_types, output_shapes=output_shapes, **config)

    def _processing(self):
        return tf.data.Dataset.from_generator(self.generator, self.param('output_types'), self.param('output_shapes'))




import rx
from fs import copy, path


def files_in_directories(fs, include_dirs, include_files, depth=0, exclude_dirs=None, exclude_files=None):
    """
    Args:
    """
    from .model.filters import FilesFilter, DirectoriesFilter
    if not isinstance(include_dirs, (list, tuple)):
        raise TypeError(
            'include_dirs should be list or tuple of str, not {}.'.format(type(include_dirs)))
    if not isinstance(include_files, (list, tuple)):
        raise TypeError(
            'include_dirs should be list or tuple of str, not {}.'.format(type(include_files)))
    results = list()    
    (DirectoriesFilter(include_dirs, exclude_dirs, depth).obv(fs)
     .flat_map(lambda p: FilesFilter(include_files, exclude_files).obv(fs, p))
     .subscribe(results.append))
    return results


def sub_directories(fs, part_of_name):
    from .model import DirectoriesFilter
    """
    A helper function for filtering directories, which is equviliant to filtering directories of pattern:
        [part_of_name, part_of_name*]
    """
    return DirectoriesFilter([part_of_name, '{n}*'.format(part_of_name)]).lst(fs)



import tensorflow as tf
from dxpy.configs import configurable
from dxpy.learn.config import config


@configurable(config, with_name=True)
def get_dist_summary(tensors, network=None, job_name='dataset', task_index=0, summary_cls_name=None, summary_config=None, name='cluster/ps/task0'):
    from dxpy.learn.train.summary_2 import SummaryWriter
    from dxpy.learn.net.zoo.srms.summ import SRSummaryWriter_v2
    with tf.device('/job:{}/task:{}'.format(job_name, task_index)):
        if summary_cls_name is None:
            sw = SummaryWriter(tensors=tensors, name=summary_config)
        elif summary_cls_name == 'srms':
            sw = SRSummaryWriter_v2(network=network, tensors=tensors, name=summary_config)

    return sw



class TaskNotFoundError(Exception):
    def __init__(self, tid=None):
        super(__class__, self).__init__(
            "Task with id: {tid} not found.".format(tid=tid))


class TaskDatabaseConnectionError(Exception):
    pass


class InvalidJSONForTask(Exception):
    def __init__(self, details):
        super(__class__, self).__init__(details)



import click
from dxpy.configs import configurable, ConfigsView
import dxpy.learn.config as config


def dataset_dist(cluster_name):
    from dxpy.learn.distribute.dataset import start_dataset_server
    start_dataset_server(name=cluster_name)



import tensorflow as tf
from dxpy.configs import configurable
from dxpy.learn.config import config
from ..graph import Graph


def global_step():
    # return tf.train.global_step(tf.get_default_session())
    # return graph().as_tensor()
    return tf.train.get_or_create_global_step()


def graph():
    if _instance is None:
        raise TypeError("Global step not created yet.")
    return _instance


def get_value():
    from dxpy.learn.session import get_default_session
    return get_default_session().run(global_step())


def set_value(value):
    graph().run('set', value)


_instance = None


class _GlobalStep(Graph):
    @configurable(config, with_name=True)
    def __init__(self, name='global_step', old_version=True, is_dist=False):
        super(__class__, self).__init__(name=name, is_dist=is_dist)
        if old_version:
            if is_dist:
                with tf.device('/job:ps/task:0'):
                    self._create_variable_and_ops()
            else:
                self._create_variable_and_ops()
        else:
            if is_dist:
                with tf.device('/job:ps/task:0'):
                    self._create_variable_and_ops_new()
            else:
                self._create_variable_and_ops_new()

    def _create_variable_and_ops_new(self):
        with tf.variable_scope('global_step', reuse=tf.AUTO_REUSE):
            gs = tf.get_variable('value', [], tf.int64)
            self.register_main_node(gs)
            self.create_placeholder_node(tf.int64, [], 'new_value')
            self.assign_op = self.as_tensor().assign(self.nodes['new_value'])
            self.register_node('setter', self.assign_op)
            self.register_task('set', self.set_value)

    def _create_variable_and_ops(self):
        self.register_main_node(tf.Variable(0, dtype=tf.int64, trainable=False,
                                            name='global_step'))
        with tf.name_scope('global_step_setter'):
            self.create_placeholder_node(tf.int64, [], 'new_value')
            self.assign_op = self.as_tensor().assign(self.nodes['new_value'])
        self.register_node('setter', self.assign_op)
        self.register_task('set', self.set_value)

    def set_value(self, feeds):
        from dxpy.learn.session import get_default_session
        get_default_session().run(self.assign_op, feed_dict={
            self.nodes['new_value']: feeds})


def create():
    global _instance
    if _instance is None:
        _instance = _GlobalStep()



import tensorflow as tf
from ..net import Net
from ..graph import NodeKeys


class SRStackCNN(Net):
    def __init__(self, name, low_image, high_image, **config):
        super().__init__(name,
                         inputs={NodeKeys.INPUT: low_image,
                                 NodeKeys.LABEL: high_image}, **config)

    @classmethod
    def _default_config(cls):
        from dxpy.collections.dicts import combine_dicts
        cfg = {
            'input_shape': [14, 14, 1],
            'label_shape': [28, 28, 1]
        }
        return combine_dicts(cfg, super()._default_config())

    @classmethod
    def _default_inputs(cls):
        return {
            NodeKeys.INPUT: {'shape': cls._default_config()['input_shape']},
            NodeKeys.LABEL: {'shape': cls._default_config()['label_shape']}
        }

    def _pre_kernel_post_inputs(self):
        from ..model.cnn.blocks import StackedConv2D
        model = StackedConv2D('stackedcnns', self.tensor(
            NodeKeys.INPUT), 10, lazy_create=True)
        self.register_child_model('sr_kernel', model)

    def _post_kernel_post_outputs(self):
        super()._post_kernel_post_outputs()
        self.register_node(NodeKeys.EVALUATE, self.tensor(NodeKeys.LOSS))

    def _kernel(self, feeds):
        from ..model.image import resize
        x = feeds[NodeKeys.INPUT]
        u = resize(x, (2, 2))
        h = self.child_model('sr_kernel')(u)
        with tf.name_scope('inference'):
            y = tf.layers.conv2d(h, 1, 3, padding='same')
        result = {NodeKeys.INFERENCE: y}
        if NodeKeys.LABEL in feeds:
            with tf.name_scope('loss'):
                l = tf.losses.mean_squared_error(feeds[NodeKeys.LABEL], y)
        result[NodeKeys.LOSS] = l
        return result





def main():
    import numpy as np
    import tensorflow as tf
    from tqdm import tqdm
    from dxpy.learn.config import config
    from dxpy.learn.scalar import create_global_scalars
    from dxpy.learn.graph import NodeKeys
    from dxpy.learn.train.summary_writer import SummaryWriter, SummaryItem
    from dxpy.learn.session import Session
    from dxpy.learn.model.tensor import PlaceHolder
    from dxpy.learn.dataset.petsino import PhantomSinograms
    from dxpy.learn.dataset.mnist import MNISTTFRecords
    from dxpy.learn.zoo.super_resolution import SRStackCNN
    from dxpy.learn.dataset.super_resolution import SuperResolutionDataset
    from dxpy.numpy_extension.visual.image import grid_view

    files = [
        '/home/hongxwing/Datas/phantom/phantom.{}.tfrecord'.format(i) for i in range(10)]
    config_sinogram = {
        'batch_size': 8,
        'files': files,
    }
    config_mnist = {
        'batch_size': 32,
        'normalization': {
            'method': 'selfminmax'
        }
    }
    config['dataset'] = {
        'sinograms': config_sinogram,
        'mnist': config_mnist
    }
    config_stacked_cnns = {
        'nb_layers': 10,
    }
    config['network'] = {
        'activation': 'relu',
        'stackedcnns': config_stacked_cnns,
    }
    create_global_scalars()

    dataset = SuperResolutionDataset('dataset', lambda: MNISTTFRecords(
        'dataset/mnist'), input_key='image', nb_down_sample=1)
    network = SRStackCNN('network', dataset['image2x'], dataset['image1x'])
    summary_loss = SummaryItem(network[NodeKeys.LOSS], 'scalar')
    summary = SummaryWriter(
        name='summary/train', tensors_to_summary={'loss': summary_loss}, path='./summary/train/')
    session = Session()
    with session.as_tensor().as_default():
        session.post_session_created()
        network.post_session_created()
        summary.post_session_created()
    with session.as_default():
        for i in range(2000):
            network.train()
            if i % 100 == 0:
                summary.summary()
        network.nodes[NodeKeys.TRAINER].decay_learning_rate(0.1)
        for i in range(2000):
            network.train()
            if i % 100 == 0:
                summary.summary()

    with session.as_default():
        imginf = network.inference(feeds={NodeKeys.INPUT: img2x})



"""
Analytical Phantom Sinogram module, provide examples with fields:
Example:
    'phantom':  scaled clean 256x256 phantom
    'sinogram': scaled clean original sinogram
    'sinogram1x': sinogram without down sampling, may apply poission sample
    'sinogram[2, 4, 8]x': down sampled with [2, 2], [4, 4], [8, 8] respectively
    'phantom[1..4]x': FBP reconstructed image (with 1e6 events) with respetive sinograms
"""
import os
import tensorflow as tf
from tables import *
from dxpy.configs import configurable
import h5py
import numpy as np
from typing import List, TypeVar
from ...config import config
from ...graph import Graph

DEFAULT_FILE_NAME = 'analytical_phantom_sinogram.h5'
NB_IMAGES = 786543


class AnalyticalPhantomSinogram(IsDescription):
    phantom = UInt16Col(shape=(256, 256))
    sinogram = UInt16Col(shape=(320, 320))
    phantom_type = UInt8Col()


class AnalyticalPhantomMultiScaleReconstruction(IsDescription):
    recon1x = UInt16Col(shape=(256, 256))
    recon2x = UInt16Col(shape=(256, 256))
    recon4x = UInt16Col(shape=(256, 256))
    recon8x = UInt16Col(shape=(256, 256))


def data_type_np(field_name):
    import tensorflow as tf
    return {
        'phantom': np.float32,
        'sinogram': np.float32,
        'phantom_type': np.int64,
        'recon1x': np.float32,
        'recon2x': np.float32,
        'recon4x': np.float32,
        'recon8x': np.float32,
    }[field_name]


def data_type_tf(field_name):
    import tensorflow as tf
    return {
        'phantom': tf.float32,
        'sinogram': tf.float32,
        'phantom_type': tf.int64,
        'recon1x': tf.float32,
        'recon2x': tf.float32,
        'recon4x': tf.float32,
        'recon8x': tf.float32,
    }[field_name]


def data_shape(field_name):
    import tensorflow as tf
    return {
        'phantom': (256, 256),
        'recon1x': (256, 256),
        'recon2x': (256, 256),
        'recon4x': (256, 256),
        'recon8x': (256, 256),
        'sinogram': (640, 320),
        'phantom_type': (),
    }[field_name]


@configurable(config.get('datasets').get('analytical_phantom_sinogram'))
def _h5files(path: str, sino_fn='analytical_phantom_sinogram.h5', recon_fn='recons.h5'):
    from dxpy.core.path import Path
    return str(Path(path) / sino_fn), str(Path(path) / recon_fn)


def _post_processing(result):
    from dxpy.medical_image_processing.projection.parallel import padding_pi2full
    from ...model.normalizer.normalizer import ReduceSum
    if 'sinogram' in result:
        result['sinogram'] = padding_pi2full(result['sinogram']).T
    for k in result:
        result[k] = result[k].astype(data_type_np(k))
        result[k] = result[k] / np.sum(result[k]) * 1e6 
    return result


def _get_example(idx, fields, h5sino, h5recon):
    result = dict()
    h5sino_keys = ['phantom', 'sinogram']
    for k in h5sino_keys:
        if k in fields:
            result[k] = h5sino.root.data[idx][k]
    h5recon_keys = ['recon1x', 'recon2x', 'recon4x', 'recon8x']
    for k in h5recon_keys:
        if k in fields:
            result[k] = h5recon.root.data[idx][k]
    return result


def dataset_generator(fields=('sinogram',), ids=None):
    if ids is None:
        ids = range(0, int(NB_IMAGES * 0.8))
        ids = list(ids)
        import random
        random.shuffle(ids)
    if isinstance(fields, str):
        fields = (fields, )
    fn_sino, fn_recon = _h5files()
    with open_file(fn_sino) as h5sino, open_file(fn_recon) as h5recon:
        for idx in ids:
            result = _get_example(idx, fields, h5sino, h5recon)
            result = _post_processing(result)
            yield result


class Dataset(Graph):
    # Statistics are calculated after fixed summation (total events) to 1e6, 
    # sinogram with minimum noise 0.4 (+0.4 to all)
    # recons with minimum noise 1.0 (+1.0 to all)
    SINO_STAT = {'mean': 4.88, 'std': 4.37}
    LOG_SINO_STAT = {'mean': 0.93, 'std': 1.46}
    RECON_STAT = {'mean': 15.25, 'std': 20.0}
    LOG_RECON_STAT = {'mean': 1.90, 'std': 1.50}

    @configurable(config, with_name=True)
    def __init__(self, name='analytical_phantom_sinogram_dataset', *,
                 batch_size=32, fields=('sinogram', ), ids=None, shuffle=True):
        if isinstance(fields, str):
            fields = (fields, )
        super().__init__(name, batch_size=batch_size,
                         ids=ids, fields=fields, shuffle=shuffle)
        dataset = self._create_dataset()
        self.dataset = self._processing(dataset)
        self._register_dataset()

    def _create_dataset(self):
        from functools import partial
        dataset_gen_partial = partial(dataset_generator,
                                      fields=self.param('fields'),
                                      ids=self.param('ids', raise_key_error=False))
        output_types = {k: data_type_tf(k) for k in self.param('fields')}
        output_shapes = {k: data_shape(k) for k in self.param('fields')}
        return tf.data.Dataset.from_generator(dataset_gen_partial, output_types, output_shapes)

    def _format_tensors(self, data):
        return {k: tf.reshape(tf.cast(data[k], data_type_tf(k)), data_shape(k)) for k in data}

    def _processing(self, dataset):
        dataset = (dataset
                   .repeat()
                   .map(self._format_tensors))
        if self.param('shuffle'):
            dataset = dataset.shuffle(1024)
        return dataset.batch(self.param('batch_size'))

    def _register_dataset(self):
        from ...utils.tensor import ensure_shape
        iterator = self.dataset.make_one_shot_iterator()
        next_element = iterator.get_next()
        self.register_main_node(next_element)
        for k in self.param('fields'):
            shape = [self.param('batch_size')] + list(data_shape(k)) + [1]
            self.register_node(k, ensure_shape(next_element[k], shape=shape))

    def data_shape(self, key):
        return data_shape(key)



# The following part is a record of dataset generator script.


# def create_dataset_from_tf_records():
#     from tqdm import tqdm
#     from dxpy.learn.dataset.config import config
#     filename = '/media/hongxwing/73f574f4-33b0-46b1-ae5d-de1eca4ef891/home/hongxwing/Workspace/Datas/phantom_sinograms_3.h5'
#     with h5py.File(filename, 'r') as fin:
#         filters = Filters(5, 'blosc')
#         h5file = open_file(str(Path(config['PATH_DATASETS']) / DEFAULT_FILE_NAME),
#                            mode="w", filters=filters)
#         table = h5file.create_table(
#             h5file.root, 'data', AnalyticalPhantomSinogram, expectedrows=fin['phantoms'].shape[0])
#         data = table.row

#         def convert_to_uint16(arr):
#             UINT16MAX = 65535
#             arr = np.maximum(arr, 0.0)
#             arr = arr / np.max(arr) * UINT16MAX
#             arr = np.minimum(arr, UINT16MAX)
#             return arr.astype(np.uint32)

#         for i in tqdm(range(fin['phantoms'].shape[0]), ascii=True):
#             data['sinogram'] = convert_to_uint16(
#                 fin['sinograms'][i, :, :320, 0])
#             data['phantom'] = convert_to_uint16(fin['phantoms'][i, :, :, 0])
#             data['phantom_type'] = np.reshape(np.array(fin['types'][i, ...],
#                                                        dtype=np.uint8), [])
#             data.append()
#     table.flush()
#     h5file.close()


# if __name__ == "__main__":
#     create_dataset_from_tf_records()



import astra
import astra.creators as ac
from ..config import config
from dxpy.configs import configurable


def _inputs_verification2d(detector, phantom_spec, sinogram):
    from ..detector.base import Detector2D
    from ..exceptions import InputVerifacationError
    if not isinstance(detector, Detector2D):
        raise InputVerifacationError("Input detector is not Detector2D, {}.")
    detector.assert_fit(sinogram)


@configurable(config.get('reconstruction'))
def reconstruction2d(sinogram, detector=None, phantom_spec=None, *, method='FBP', iterations=1, nb_sensors=None, nb_views=None, sensor_width=None, image_shape=None):
    """
    Args:
        detector: Detector specifics,
        phantom_spec: phantom specifics
        sinogram: 2-dimensional ndarray.
    """
    from ..detector.parallel import Detector2DParallelRing
    from ..phantom.base import Phantom2DSpec
    if detector is None:
        detector = Detector2DParallelRing(
            nb_sensors=nb_sensors, sensor_width=sensor_width, nb_views=nb_views)
    if phantom_spec is None:
        phantom_spec = Phantom2DSpec(image_shape)
    _inputs_verification2d(detector, phantom_spec, sinogram)
    vol_geom = ac.create_vol_geom(phantom_spec.shape)
    proj_geom = ac.create_proj_geom('parallel',
                                    detector.sensor_width,
                                    detector.nb_sensors,
                                    detector.views)
    proj_id = ac.create_projector('linear', proj_geom, vol_geom)
    rid, result = ac.create_reconstruction(method,
                                           proj_id,
                                           sinogram,
                                           iterations)
    astra.data2d.clear()
    astra.projector.clear()
    astra.algorithm.clear()
    return result


class Reconstructor2DParallel:
    @configurable(config.get('projection'))
    def __init__(self, detector, phantom_spec, *, method='FBP_CUDA', projection_model='linear'):
        vol_geom = ac.create_vol_geom(phantom_spec.shape)
        proj_geom = ac.create_proj_geom('parallel',
                                        detector.sensor_width,
                                        detector.nb_sensors,
                                        detector.views)
        self.proj_id = ac.create_projector(
            projection_model, proj_geom, vol_geom)
        self.sino_id = astra.data2d.create('-sino', proj_geom)
        self.method = method

    def reconstruct(self, sinogram, iterations=1):
        astra.data2d.store(self.sino_id, sinogram)
        iid, result = ac.create_reconstruction(self.method,
                                               self.proj_id,
                                               sinogram,
                                               iterations)
        astra.data2d.delete(iid)
        return result

    def clear(self):
        astra.data2d.delete(self.phan_id)
        astra.projector.delete(self.proj_id)



""" base and core utilities of dxpy """
import json

class JSONSerializable:
    # @classmethod
    def to_json(self):
        return json.dumps(self.__dict__)

    @classmethod
    def from_json(cls, dct):
        return cls(**json.loads(dct))

def say_hello():
    print("Hello from dxpy.")


import tensorflow as tf
from ..base import Model


class StackDense(Model):
    def __init__(self, name, input_, filters):
        super().__init__(name, inputs={'input': input_}, filters=filters)

    def _kernel(self, feeds):
        x = feeds['input']
        for i, f in enumerate(self.param('filters')):
            if i < len(self.param('filters')) - 1:
                activ = tf.nn.relu
            else:
                activ = None
            x = tf.layers.dense(x, f, activation=activ)
        return x



from pprint import pprint

import numpy as np
import tensorflow as tf
import logging
from dxpy.learn.dataset.zoo.sraps import AnalyticalPhantomSinogramDatasetForSuperResolution
from dxpy.learn.graph import NodeKeys
from dxpy.learn.net.zoo.srms import SRMultiScale
from dxpy.learn.session import Session
from dxpy.learn.train.summary import SummaryWriter
from dxpy.learn.utils.general import pre_work
from dxpy.learn.model.cnn.super_resolution import SuperResolutionMultiScale, BuildingBlocks
from tqdm import tqdm
from dxpy.learn.config import config
import yaml
from .summ import SRSummaryWriter
from dxpy.learn.run.train import train

def train_defs(external_configs):
    dataset = AnalyticalPhantomSinogramDatasetForSuperResolution()
    inputs = {}
    for i in range(dataset.param('nb_down_sample') + 1):
        inputs.update({'input/image{}x'.format(2**i): dataset['input/image{}x'.format(2**i)]})
        if external_configs['use_noise_label']:
            inputs.update({'label/image{}x'.format(2**i): dataset['input/image{}x'.format(2**i)]})
        else:
            inputs.update({'label/image{}x'.format(2**i): dataset['label/image{}x'.format(2**i)]})
            if dataset.param('image_type') == 'image':
                inputs['label/image1x'] = dataset['label/phantom']
    network = SRMultiScale(inputs, name='network')
    summary = SRSummaryWriter(name='summary/train', network=network)
    return network, summary
    
def main():
    train(train_defs)


if __name__ == "__main__":
    main()


def type_error_message(required_type, got_type, prefix=""):
    if not isinstance(prefix, str):
        from .common import DXPYTypeError
        raise DXPYTypeError(str, prefix, "Arg prefix of type_error_")
    if not prefix == "":
        return "{} is required to be {}, got {}.".format(prefix, required_type, got_type)



# """ Config utility class """
# from functools import getfullargspec
# import json

# def with_config(func):
#     """ Add support to read keywargs from JSON file

#     Add a preserved keyword `config_filenames` for JSON file names to read.
#     """
#     @wraps(func)
#     def wrapper(*args, config_filenames=None, **kwargs):
#         if isinstance(config_filenames, str):
#             config_filenames = [config_filenames]
#         if config_filenames is None:
#             config_filenames = []
#         paras = getfullargspec(func)
#         json_dicts = []
#         for fn in config_filenames:
#             with open(fn, 'r') as fin:
#                 json_dicts.append(json.load(fin))
#         if paras.defaults is not None:
#             nb_def = len(paras.defaults)
#         else:
#             nb_def = 0
#         def_args = paras.defaults
#         if def_args is None:
#             def_args = []
#         def_keys = paras.args[-nb_def:]
#         def_dict = {k: v for k, v in zip(def_keys, def_args)}
#         for k in paras.args:
#             v = config_from_dicts(k, [kwargs] + json_dicts + [def_dict])
#             if v is not None:
#                 kwargs.update({k: v})
#         kwargs.update({'config_filenames': config_filenames})
#         kwargs.pop('config_filenames')
#         return func(*args, config_filenames=config_filenames, **kwargs)
#     return wrapper


# class Config:
#     """ Base class of configs """

#     def __init__(self,
#                  config_file=None,
#                  **kwargs):
#         self._configs = dict(**kwargs)
#         self._config_file = config_file

#     def load_config_file(self, file_name = None : str):
#         """ Load JSON config file, update config_file property if a new file_name is given. 
#         Inputs:
#             file_name: [Optional] JSON config file path.
#         Returns:
#             None
#         Raises:
#             //TODO: Complete
#         """
#         if file_name is None:
#             file_name = self._config_file
#         else:
#             self._config_file = file_name
#         with open(file_name, 'r') as fin:
#             self._configs = json.load(fin)

# """ Config utility class """
import json
import inspect
from functools import wraps
from pathlib import Path
from collections import ChainMap

from xlearn.utils.collections import AttrDict


def __parse_config_args(func, *args, config_files=None, **kwargs):
    """ Analysic func args and inputs/JSON file.

    Inputs:

        * args: tuple of positional arguments,                
        * config_files: filename or list of filenames of JSON file(s), *which may
          or may not present in original function's argument list*.
        * kwargs: dict of keyword arugments.

    Returns:
        1. a tuple, positional arguments,        
        2. a dict, keyword arugments.

    Raises:

    """
    # TODO: refine doc, add test.
    if isinstance(config_files, str):
        config_files = [config_files]
    if config_files is None:
        config_files = []
    json_dicts = []
    for fn in config_files:
        with open(fn, 'r') as fin:
            json_dicts.append(json.load(fin))
    sig = inspect.signature(func)
    kwargs['config_files'] = config_files
    kwargs = ChainMap(*([kwargs] + json_dicts))
    func_arg_keys = [k for k in sig.parameters]
    poped_keys = []
    func_arg_keys_tmp = list(func_arg_keys)
    for a, k in zip(args, func_arg_keys_tmp):
        poped_keys.append(func_arg_keys.pop(func_arg_keys.index(k)))
    kwargs_fine = {k: kwargs.get(k)
                   for k in func_arg_keys if kwargs.get(k) is not None}
    ba = sig.bind(*args, **kwargs_fine)
    ba.apply_defaults()
    return ba.args, ba.kwargs, poped_keys


def json_config(func):
    """ Add auto jason kwargs to function, by adding a preserved keyword 'config_files' """
    @wraps(func)
    def wrapper(*args, config_files=None, **kwargs):
        f_args, f_kwargs, _ = __parse_config_args(
            func, *args, config_files=config_files, **kwargs)
        return func(*f_args, **f_kwargs)
    return wrapper



class PathTypeError(TypeError):
    def __init__(self, wrong_type):
        super().__init__("Can not convert {} to Path.".format(wrong_type))


class PathValueError(ValueError):
    def __init__(self, wrong_value):
        super().__init__("Can not parst {} as Path.".format(wrong_value))



""" File objects
"""
# TODO: Add yaml support

from abc import ABCMeta, abstractmethod
from dxpy.file_system.path import Path


class FileError(Exception):
    pass


class DxlFileExistsError(FileError, FileExistsError):
    pass


class FileNotFoundOrWrongTypeError(FileError, FileNotFoundError):
    def __init__(self, path, ftype):
        self.path = path
        self.ftype = ftype

    def __str__(self):
        msg = r"File {path} with type {ftype} not found."
        return msg.format(path=self.path, ftype=self.ftype)


# class Attrs:
#     @staticmethod
#     def get_attr_by_name(name):
#         for x in dir(Attrs):
#             x = getattr(Attrs, x)
#             if not type(x) == type(Attrs.FileAttrsABC):
#                 continue
#             if issubclass(x, Attrs.FileAttrsABC) and x.name == name:
#                 return x
#         raise NotValidAttrName(name)

#     @staticmethod
#     def get_attr_by_names(names):
#         return {Attrs.get_attr_by_name(n) for n in names}

#     @staticmethod
#     def get_attrs_by_path(path, file_system):
#         if isinstance(path, str):
#             path = Path(path)
#         if not FileSystem.exists(path):
#             raise FileNotFoundOrWrongTypeError(path.abs)
#         result = set()
#         for x in dir(Attrs):
#             x = getattr(Attrs, x)
#             if type(x) == type(Attrs.FileAttrsABC) and issubclass(x, Attrs.FileAttrsABC):
#                 result = result.union(x.test(path))
#         return result

#     class File(FileAttrsABC):
#         name = 'FILE'

#         @staticmethod
#         def test(path):
#             return {Attrs.File, } if FileSystem.is_file(path) else set()

#     class Dir(FileAttrsABC):
#         name = 'DIR'

#         @staticmethod
#         def test(path):
#             return {Attrs.Dir, } if FileSystem.is_dir(path) else set()

#     class Text(FileAttrsABC):
#         name = 'TEXT'

#         @staticmethod
#         def test(path):
#             return {Attrs.Text, } if path.suffix in {'.txt'} else set()

#     class Exec(FileAttrsABC):
#         name = 'EXEC'

#     class ShellScript(Exec):
#         name = 'SHELLSCRIPT'

#         @staticmethod
#         def test(path):
#             return {Attrs.ShellScript, } if path.suffix == '.sh' else set()

#     class WinExe(Exec):
#         name = 'WINEXE'

#         @staticmethod
#         def test(path):
#             return {Attrs.WinExe, } if path.suffix == '.exe' else set()

#     class Bin(FileAttrsABC):
#         name = 'BIN'

#     class NumpyData(Bin):
#         name = 'NUMPYDATA'

#         @staticmethod
#         def test(path):
#             return {Attrs.NumpyData, } if path.suffix in ['.npy', '.npz'] else set()

#     class NPY(NumpyData):
#         name = 'NPY'

#         @staticmethod
#         def test(path):
#             return {Attrs.NPY, } if path.suffix in ['.npy'] else set()

#     class NPZ(NumpyData):
#         name = 'NPZ'

#         @staticmethod
#         def test(path):
#             return {Attrs.NPZ, } if path.suffix in ['.npz'] else set()


# class FileAttrs:
#     def __init__(self, attrs=None):
#         if attrs is None:
#             self.attrs = set()
#             return
#         if isinstance(attrs, FileAttr):
#             self.attrs = attrs.attrs
#             return
#         if not isinstance(attrs, (list, tuple, set)):
#             attrs = {attrs}

#     def add(self, attr):
#         self.attrs.add(attr)

#     def __str__(self):
#         return '[' + ','.join(self.info) + ']'

#     @property
#     def brief(self):
#         return [x.name for x in self.attrs]

#     @property
#     def detail(self):
#         return self.brief

#     def hasattr(self, attr):
#         return any(issubclass(x, attr) for x in self.attrs)

#     @property
#     def is_exe(self):
#         return self.hasattr(Attrs.Exec)

#     @property
#     def is_dir(self):
#         return self.hasattr(Attrs.Dir)

#     def check(self, path, file_system):
#         for aAttr in self.attrs:
#             if not aAttr.test(path, file_system):
#                 raise FileAttrError(path, self.attrs, aAttr)
#         return None


class FileAbstract(metaclass=ABCMeta):
    """ File class focus on existance.
    """

    def __init__(self, path, file_system):
        self.path = Path(path)
        self.fs = file_system
        if not self.check(self.path, self.fs):
            raise FileNotFoundOrWrongTypeError(self.path, self.ftype)

    @staticmethod
    def check(path, file_system):
        if isinstance(path, str):
            path = Path(path)
        return file_system.exists(path)

    def copy_to(self, path):
        result = self.fs.copy(path, self.path)
        return result

    def move_to(self, path):
        result = self.fs.move(path, self.path)
        return result

    @abstractmethod
    def update(self, content):
        return NotImplementedError

    @abstractmethod
    def delete(self):
        raise NotImplementedError

    @property
    def ftype(self):
        return {'FileAbstract'}

    @property
    def brief(self):
        """ A JSON serilizable info dict.
        """
        types = [str(x) for x in self.ftype]
        types.sort()
        return {
            'path': self.path.abs,
            'name': self.path.name,
            'type': types}

    @property
    def detail(self):
        result = self.brief
        result.update({
            'parent': self.path.parent,
            'parts': self.path.parts(),
        })
        return result


class Directory(FileAbstract):
    def __init__(self, path, file_system, factory=None):
        if factory is None:
            factory = FileFactory
        FileAbstract.__init__(self, path, file_system)
        self.factory = factory
        if not self.check(self.path, self.fs):
            raise FileNotFoundOrWrongTypeError(self.path, self.ftype)

    @staticmethod
    def check(path, file_system):
        if isinstance(path, str):
            path = Path(path)
        return file_system.Directory.check(path)

    def delete(self):
        self.fs.Directory.delete(self.path)
        return None

    def update(self):
        return self

    @property
    def ftype(self):
        result = super().ftype
        result.add('Directory')
        return result

    @property
    def detail(self):
        children_path = self.fs.Directory.glob(self.path)
        children_brief = [FileFactory.get_file(
            cp, self.fs) for cp in children_path]

        result = super().detail
        result.update({
            "children": children_brief
        })
        return result


class NormalFile(FileAbstract):
    def __init__(self, path, file_system):
        FileAbstract.__init__(path, file_system)
        if not self.check(self.path, self.fs):
            raise FileNotFoundOrWrongTypeError(self.path, self.ftype)

    @staticmethod
    def check(path, file_system):
        return file_system.File.check(path)

    def delete(self):
        self.fs.File.delete(self.path)

    def update(self):
        return self

    @property
    def ftype(self):
        result = super().ftype
        result.add('NormalFile')
        return result


class Exec(NormalFile):
    pass


class FileFactory:
    @staticmethod
    def get_file(path, file_system):
        fa = FileAttr().set_attrs(path)
        if fa.hasattr(Attrs.Dir):
            return Directory(path)
        if fa.hasattr(Attrs.File):
            return NormalFile(path)

    @staticmethod
    def create_file(path, file_system, cls):
        pass



"""
Use case:
a function with parameters foo(**kw, name),

if is defined like:
@configurable()
def foo(xx=None, yy=None, zz=None, name='default_name'):
    pass
Will try to load config by c['defautl_name']

if is defined like:
@configurable('conf_name')
def foo(**kw):
    pass
Will try to load config by c['conf_name']
"""
def configurable(name=None):
    pass



from dxpy.learn.dataset.api import get_dataset
from dxpy.learn.net.api import get_network
from dxpy.learn.session import SessionDist
from tqdm import tqdm
from dxpy.learn.graph import NodeKeys
import time
import arrow
from dxpy.learn.scalar import global_step
import tensorflow as tf


def main(task='train', job_name='worker', task_index=0, cluster_config='cluster.yml'):
    from dxpy.learn.distribute.cluster import get_cluster_spec, get_server, get_nb_tasks
    from dxpy.learn.distribute.dataset import get_dist_dataset
    from dxpy.learn.distribute.ps import get_dist_network as get_ps_network
    from dxpy.learn.distribute.worker import apply_dist_network
    from dxpy.learn.distribute.summary import get_dist_summary
    from dxpy.learn.distribute.worker import get_dist_network as get_worker_network
    from dxpy.learn.utils.general import pre_work
    config = tf.ConfigProto()
    config.gpu_options.allow_growth = True
    cluster = get_cluster_spec(cluster_config, job_name=None)
    server = get_server(cluster, job_name, task_index, config=config)
    dataset = get_dist_dataset(name='cluster/dataset/task0')

    if job_name == 'dataset':
        server.join()
        return
    elif job_name == 'ps':
        server.join()
        return
    elif job_name in ['worker', 'summary']:
        with tf.device(tf.train.replica_device_setter(worker_device="/job:worker/task:{}".format(task_index), cluster=cluster)):
            pre_work()
            network = get_network(name='network/srms', dataset=dataset)
            result = network()
        if job_name == 'worker':
            hooks = [tf.train.StepCounterHook()]
            with tf.train.MonitoredTrainingSession(master=server.target,
                                                   is_chief=(task_index == 0),
                                                #    is_chief=True,
                                                   checkpoint_dir="./save",
                                                   config=config,
                                                   hooks=hooks) as sess:
                from dxpy.learn.session import set_default_session
                set_default_session(sess)
            # sess = SessionDist(target=server.target)
                # with sess._sess.as_default():
                for _ in tqdm(range(10000000000000), ascii=True):
                    network.train()
    if job_name == 'summary':
        name = 'cluster/summary/task{}'.format(task_index)
        result = apply_dist_network(
            network=network, dataset=dataset, name=name)
        result.update(dataset.nodes)
        sw = get_dist_summary(tensors=result, network=network, name=name)
        sess = SessionDist(target=server.target)
        with sess.as_default():
            sw.post_session_created()
            while True:
                sw.auto_summary()
        return



from .base import ScalarVariable
from .global_global_step import global_step
from .global_global_step import get_value as current_step
from .global_keep_prob import keep_prob
from .global_batch_size import batch_size
from .utils import create_global_scalars



from collections import UserDict


class Config(UserDict):
    version = 0.1

    def __init__(self, restful_name='fs'):
        from fs.osfs import OSFS
        super(__class__, self).__init__()
        self.restful_name = restful_name
        self.default_fs = OSFS


c = Config()



import numpy as np

def psnr(target, label, max_value=None, tor=1e-7):
    shape_check(target, label)
    if max_value is None:
        max_value = np.max([target, label])
    err_square = np.square(target - label)
    err_sum = np.sum(err_square, axis=(1, 2, 3))
    err_sum[err_sum < tor] = tor
    psnr_value = 10.0 * (1.0 + np.log(np.prod(target.shape[1:]) * max_value * max_value / err_sum))
    return psnr_value



import pathlib
import shutil
from dxpy.file_system.path import Path


class FileSystem:
    @staticmethod
    def exists(path):
        if isinstance(path, str):
            path = Path(path)
        return pathlib.Path(path.abs).exists()

    @staticmethod
    def copy(path_target, path_source):
        raise NotImplementedError

    @staticmethod
    def move(path_target, path_source):
        raise NotImplementedError

    class File:
        @staticmethod
        def check(path):
            return pathlib.Path(path.abs).is_file()

        @staticmethod
        def create(path):
            # Get a emtpy file by FileFactory
            # Update file with content
            pathlib.Path(path.abs).touch()

        @staticmethod
        def delete(path):
            pathlib.Path(path.abs).unlink()

    class Directory:
        @staticmethod
        def check(path):
            return pathlib.Path(path.abs).is_dir()

        @staticmethod
        def create(path, parents=False):
            if isinstance(path, str):
                path = Path(path)
            return pathlib.Path(path.abs).mkdir(parents=parents)

        @staticmethod
        def delete(path):
            shutil.rmtree(path.abs)

        @staticmethod
        def glob(path, pattern='*'):
            return [str(x.absolute()) for x in pathlib.Path(path.abs).glob(pattern)]



import click


class CLI(click.MultiCommand):
    """
    Task database utils.
    """
    commands = {'start': None}

    def __init__(self):
        super(__class__, self).__init__(name='db', help=__class__.__doc__)

    def list_commands(self, ctx):
        return sorted(self.commands.keys())

    def get_command(self, ctx, name):
        from . import cmds
        if name in self.commands and self.commands[name] is None:
            self.commands[name] = getattr(cmds, name)
        return self.commands.get(name)


cli = CLI()
